"""

    ElutionMapper.py

        Optimization of mapping between UV-absorbance and Xray-scattering

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import copy
import numpy                as np
from scipy.interpolate      import UnivariateSpline
from scipy import optimize
from scipy.spatial import distance
import logging
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve, PEAK_INFO_FIND_RATIO
from molass_legacy.Elution.CurveUtils      import get_probably_corresponding_index
from molass_legacy._MOLASS.SerialSettings         import ( get_setting, reset_setting, set_setting,
                                        UV_BASE_NO, UV_BASE_CONST,
                                        UV_BASE_STANDARD, UV_BASE_SHIFTED,
                                        XRAY_BASE_NO, XRAY_BASE_CONST,
                                        INTEGRAL_BASELINE
                                        )
from .MappingParams import set_mapper_opt_params, get_mapper_opt_params, MappedInfo
from molass_legacy.SerialAnalyzer.ScatteringBaseCorrector import compute_baseline_using_LPM_impl
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from molass_legacy.KekLib.CanvasDialog import CanvasDialog
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception
# from molass_legacy.SerialAnalyzer.AnimObjects import AnimMapper
from .PeakMapper import PeakMapper
from .SingleComponent       import SingleComponent, PEAK_EVAL_RANGE_RATIO
from molass_legacy.UV.XrayProportional import make_proportional_mapping_info_impl
import molass_legacy.KekLib.DebugPlot as plt

# EQUALIZE_OUTSIDE_OF_FC  = True
ZEROS_TO_OUTSIDE_OF_FC  = True
MAPPED_RATIO_ALLOW      = 0.5       # > 0.21 for 20180225, > 0.35 for 20160628
ALLOWANCE_PERCENT       = 10        # > 5 for 20161119/Kosugi3a
SCALE_MIN_RATIO         = 0.7
SCALE_MAX_RATIO         = 1.4
AJUSTMENT_ITERATION     = 2
XRAY_INTERCEPT_PERCENT  = 5
USE_TOTAL_STD_DIFF      = True
USE_WEIGHTED_DIFF       = True
STD_DIFF_SCALE          = 10        # this value is for the case where USE_TOTAL_STD_DIFF == True
VALLEY_DIFF_WEIGHT      = 0.5
MAPPING_RANGE_RATIO     = PEAK_INFO_FIND_RATIO
MAPPING_RANGE_DELTA     = 0.03      # adjustment for cases such as 20161202
USE_SAVED_INFO_TO_GET_RANGES    = False
REFINE_MAPPING_RANGES   = True
USE_SIMPLE_SCALE_FOR_VALLEY = False
SMALL_VALUE_LIMIT       = 1e-5
USE_SAME_RATIO_BOUNDARY = True
ACCEPTABLE_STD_DIFF     = 1.0       # 0.63 for Kosugi3a
REQUIRE_HELPER_STD_DIFF = 1.0
VERY_SMALL_RATIO        = 1e-5
USE_AREA_SCALING        = True
PENALTY_KEEP_POS_VALUE  = 1e-5
PENALTY_KEEP_POS_RATIO  = 0.03
USE_SIMPLE_MAPPING      = True

TO_BE_REMOVED   = False

"""
refactoring memo

    optimize
        optimize_mapping

        adjust_baselines
            adjust_uv_baseline_only
                minimize
            adjust_xray_baseline_only
                minimize
            adjust_both_baselines
                minimize
"""

class ElutionMapper:
    def __init__(self, serial_data, input_sd, pre_recog, callbacks=None):
        self.logger = logging.getLogger( __name__ )

        assert serial_data.orig_info is None

        self.orig_sd = serial_data
        self.input_sd = input_sd
        self.pre_recog = pre_recog
        self.initialize(self.orig_sd)

        self.num_iterations = get_setting( 'correction_iteration' )
        self.use_xray_conc  = get_setting('use_xray_conc')
        self.use_mtd_conc  = get_setting('use_mtd_conc')
        self.enable_lrf_baseline = get_setting('enable_lrf_baseline')
        self.baseline_corrected_copy = None
        self.sd_update_callback = None
        self.optimize_callback = None
        self.x_curve_update_callback = None
        if callbacks is not None:
            self.sd_update_callback = callbacks[0]
            self.optimize_callback = callbacks[1]
            self.x_curve_update_callback = callbacks[2]

        self.logger.info('mapper has been created with conc_type %d.' % self.get_conc_type() )

        if False:
            fig = plt.figure(figsize=(12,6))
            fig.suptitle("ElutionMapper.__init__")
            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)
            ax1.plot(self.a_vector)
            ax1.plot(self.a_spline(self.a_x))
            ax2.plot(self.x_vector, color='orange')
            fig.tight_layout()
            plt.show()

    def set_sd_update_callback(self, callback):
        self.sd_update_callback = callback

    def set_optimize_callback(self, callback):
        self.optimize_callback = callback

    def set_x_curve_update_callback(self, callback):
        self.x_curve_update_callback = callback

    def initialize(self, sd):
        self.serial_data = sd.get_copy()
        self.top_x_pair = sd.get_original_peak_top_x_pair()
        self.orig_info  = sd.orig_info
        self.absorbance = sd.absorbance
        if self.absorbance is None:
            self.logger.error("serial_data.absorbance is None. state the recipe here.")
            raise RuntimeError("Unprepared Serial Data")

        self.sd_xray_curve = sd.xray_curve
        self.a_vector   = self.absorbance.a_vector
        self.a_x        = np.arange( len(self.a_vector) )
        self.a_spline   = UnivariateSpline( self.a_x, self.a_vector, s=0, ext=3 )
        self.x_vector   = sd.ivector
        self.x_x        = np.arange( len(self.x_vector) )
        self.x_spline   = UnivariateSpline( self.x_x, self.x_vector, s=0, ext=3 )
        self.pre_recog  = sd.pre_recog
        self.feature_mapped = False

        for info_name in ['decomp_editor_info', 'range_editor_info']:
            info_value = get_setting(info_name)
            if info_value is not None:
                reset_setting(info_name)
                self.logger.info("cleared memorized %s", info_name)

    def get_conc_type( self ):
        # TODO: unify
        if self.use_xray_conc:
            type_ = 1
        elif self.use_mtd_conc:
            type_ = 2
        else:
            type_ = 0
        return type_

    def set_opt_params(self, mapping_params):
        self.opt_params = mapping_params
        btype = mapping_params['uv_baseline_type']
        self.absorbance.compute_base_curve(self.pre_recog, btype)
        self.optimize(opt_params=self.opt_params)
        set_mapper_opt_params(self.opt_params)

    def optimize( self, opt_params=None, curve_make_only=False, sync_options=None, helper_info=None, apply_patch=False ):
        self.initialize(self.orig_sd)   # possibly required to ensure consistency in multiple optimization; making sure

        self.uniformly_scaled_vector = None
        self.sci_list = None
        self.cd_degrees = None
        self.A_simple = None

        if self.optimize_callback is not None:
            self.optimize_callback()

        if opt_params is None:
            opt_params = get_mapper_opt_params()

        self.logger.info( 'mapping optimization with parameters: ' + str(opt_params) )

        conc_type = self.get_conc_type()
        if conc_type == 1:
            make_proportional_mapping_info_impl(self, opt_params)
            return
        elif conc_type == 2:
            import Microfluidics
            from MicrofluidicMapping    import make_microfluidic_mapping_info_impl
            make_microfluidic_mapping_info_impl(self, opt_params)
            return
        else:
            pass

        self.opt_params = opt_params
        self.sync_options = sync_options
        self.make_elution_curves( opt_params, helper_info )

        self.x_curve_y_adjusted = self.x_curve.y    # for convenience for ElutinMapperCanvas.draw_mapped

        self.ensure_peak_consistency()

        if curve_make_only:
            return

        self.compute_initial_params()

        self.in_uv_adjustment_mode = False
        self.in_xray_adjustment_mode = False

        try:
            self.optimize_mapping() 
        except:
            log_exception(self.logger, "optimize_mapping failure")
            self.resort_to_simpler_solution()

        self.adjust_baselines( opt_params )

        self.set_peak_eval_ranges()
        self.compute_std_diff()

        if False:
            A_avg   = np.average( [ result[0] for result in self.opt_results ] )
            B_avg   = np.average( [ result[1] for result in self.opt_results ] )
            self.check_mapping_normality( A_avg, B_avg, std_diff=self.std_diff )

        self.scomp      = SingleComponent(self)

        self.debug_log_results()

    def debug_log_results(self):
        self.logger.info("optimize: map_params=%s, scale_params=%s", str(self.map_params), str(self.scale_params))

    def resort_to_simpler_solution(self):
        pass

    def make_elution_curves(self, opt_params, helper_info, debug=False):
        a_curve_temp = self.make_a_curve( opt_params, helper_info )
        x_curve_temp = self.make_x_curve( opt_params )

        self.peak_mapper = pm = PeakMapper( a_curve_temp, x_curve_temp, pre_recog=self.pre_recog, debug=False )
        mapped_curves   = pm.get_mapped_curves()
        self.a_curve    = mapped_curves[0]
        self.x_curve    = mapped_curves[1]
        if self.x_curve_update_callback is not None:
            self.x_curve_update_callback(self.x_curve)
        self.x_curve.get_emg_peaks(logger=self.logger)

        if debug:
            # note that mapped_curves hve been corrected

            def plot_curve(ax, curve):
                ax.plot(curve.x, curve.y)

            curves = self.orig_sd.get_elution_curves()
            with plt.Dp():
                fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))
                fig.suptitle("make_elution_curves debug")
                plot_curve(axes[0,0], curves[0])
                plot_curve(axes[0,1], curves[1])
                plot_curve(axes[1,0], self.mapped_curves[0])
                plot_curve(axes[1,1], self.mapped_curves[1])
                fig.tight_layout()
                plt.show()

        if self.sync_options is None:
            self.feature_mapped = pm.feature_mapped
            self.manual_sync = False
        else:
            if pm.feature_mapped and self.sync_options == 1:
                self.feature_mapped = True
            else:
                self.feature_mapped = False

            if self.sync_options == 2:
                self.manual_sync = True
            else:
                self.manual_sync = False

    def make_a_curve(self, opt_params, helper_info=None, flow_change_exist=False, devel=True):
        absorbance  = self.absorbance
        uv_baseline_opt = opt_params['uv_baseline_opt']

        if uv_baseline_opt == 0:
            self.a_base = np.zeros( len(self.a_vector) )
        else:
            uv_baseline_type = opt_params['uv_baseline_type']
            self.logger.info("make_a_curve: uv_baseline_type=%d", uv_baseline_type)

            if uv_baseline_type == 1:
                self.a_base = absorbance.get_standard_elution_base()
            elif uv_baseline_type == 4:
                a_base_ok = False
                if self.enable_lrf_baseline:
                    if devel:
                        from importlib import reload
                        import Trimming.UvBaseSolver
                        reload(Trimming.UvBaseSolver)
                    from molass_legacy.Trimming.UvBaseSolver import get_lrf_elution_base
                    try:
                        self.a_base = get_lrf_elution_base(self.input_sd, self.pre_recog)
                        a_base_ok = True
                    except:
                        log_exception(self.logger, "get_lrf_elution_base failed: ", n=10)
                if not a_base_ok:
                    self.a_base = absorbance.get_shifted_elution_base()
            elif uv_baseline_type == 5:
                self.a_base = absorbance.get_integral_basecurve()
            else:
                assert False

        ac_vector       = self.a_vector - self.a_base

        debug = False

        if debug:
            # import traceback
            # traceback.print_stack()
            ax  = plt.gca()
            ax.cla()
            ax.set_title( "a_vector and a_base" )
            ax.plot( self.a_vector )
            ax.plot( self.a_base, color='red' )
            plt.tight_layout()
            plt.show()

        if flow_change_exist:
            self.flow_changes = absorbance.get_real_flow_changes()

            if helper_info is None:
                fc  = self.flow_changes
            else:
                fc  = helper_info[2]
            # flatten outside of flow_changes
            if ZEROS_TO_OUTSIDE_OF_FC:
                if fc[0] is not None:
                    ac_vector[0:fc[0]]  = 0
                if fc[1] is not None:
                    ac_vector[fc[1]+1:] = 0
            else:
                if fc[0] is not None:
                    ac_vector[0:fc[0]]  = ac_vector[ fc[0] ]
                if fc[1] is not None:
                    ac_vector[fc[1]+1:] = ac_vector[ fc[1] ]
        else:
            # this is a temp fix.
            # better be included in the above case.
             self.flow_changes = [None, None]

        max_y = None if self.orig_info is None else self.orig_info[0]
        orig_top_x = None if self.top_x_pair is None else self.top_x_pair[0]
        a_curve_temp = ElutionCurve( ac_vector, max_y=max_y, orig_top_x=orig_top_x )

        if debug:
            from CanvasDialog   import CanvasDialog
            from molass_legacy.SerialAnalyzer.ElutionCurve   import proof_plot
            def plot_func( fig, parent ):
                proof_plot( a_curve_temp, parent, fig )
            dialog = CanvasDialog( "Debug: a_curve", adjust_geometry=True )
            dialog.show( plot_func, figsize=(16, 8), parent_arg=True )

        return a_curve_temp

    def make_x_curve( self, opt_params ):
        xray_baseline_opt = opt_params[ 'xray_baseline_opt' ]
        if xray_baseline_opt == XRAY_BASE_NO:
            self.x_base = np.zeros( len(self.x_vector) )
        elif xray_baseline_opt == XRAY_BASE_CONST:
            self.x_base = self.compute_xray_baseline(opt_params)
        else:
            assert False

        xc_vector       = self.x_vector - self.x_base
        max_y = None if self.orig_info is None else self.orig_info[1]
        orig_top_x = None if self.top_x_pair is None else self.top_x_pair[1]
        x_curve_temp    = ElutionCurve( xc_vector, max_y=max_y, orig_top_x=orig_top_x )
        return x_curve_temp

    def ensure_peak_consistency( self, check_raw_info=False ):
        # for peak_info
        try:
            allowance = int( len(self.x_curve.x) * ALLOWANCE_PERCENT/100 )
            remove_info = self.check_peak_concictency( self.a_curve.peak_info, self.x_curve.peak_info, allowance )
            if len( remove_info[0] ) > 0:
                self.a_curve.remove_peaks( remove_info[0] )
                self.logger.warning( 'removed peaks ' + str(remove_info[0]) + ' from a_curve' )

            if len( remove_info[1] ) > 0:
                self.x_curve.remove_peaks( remove_info[1] )
                self.logger.warning( 'removed peaks ' + str(remove_info[1]) + ' from x_curve' )
        except:
            etb = ExceptionTracebacker()
            self.logger.error( 'failed in removing peaks:\n' + str(etb) )
            # this exception should cause the following exception 

        a_num_peaks = len( self.a_curve.peak_info )
        b_num_peaks = len( self.x_curve.peak_info )
        if a_num_peaks != b_num_peaks:
            raise RuntimeError( 'Numbers of peaks are inconsistent: %d != %d' % ( a_num_peaks, b_num_peaks ),
                                [ 1, a_num_peaks, b_num_peaks ] )

        a_num_boundaries = len( self.a_curve.boundaries )
        b_num_boundaries = len( self.x_curve.boundaries )
        if a_num_boundaries != b_num_boundaries:
            raise RuntimeError( 'Numbers of boundaries are inconsistent: %d != %d' % ( a_num_boundaries, b_num_boundaries ),
                                [ 1, a_num_boundaries, b_num_boundaries ] )

        if check_raw_info:

            # for peak_info_raw with smaller allowance to cope with such cases os 20161006/OA01
            a_peak_info_raw = self.a_curve.raw_info[PEAK_INFO_RAW]
            x_peak_info_raw = self.x_curve.raw_info[PEAK_INFO_RAW]

            try:
                allowance_raw = allowance/4     # allowance/2 is not acceptable for 20160628
                remove_info = self.check_peak_concictency( a_peak_info_raw, x_peak_info_raw, allowance_raw )
                if len( remove_info[0] ) > 0:
                    self.a_curve.remove_raw_peaks( remove_info[0] )
                    self.logger.warning( 'removed raw peaks ' + str(remove_info[0]) + ' from a_curve' )

                if len( remove_info[1] ) > 0:
                    self.x_curve.remove_raw_peaks( remove_info[1] )
                    self.logger.warning( 'removed raw peaks ' + str(remove_info[1]) + ' from x_curve' )
            except:
                etb = ExceptionTracebacker()
                self.logger.error( 'failed in removing raw peaks:\n' + str(etb) )
                # this exception should cause the following exception 

            a_peak_info_raw = self.a_curve.raw_info[PEAK_INFO_RAW]
            x_peak_info_raw = self.x_curve.raw_info[PEAK_INFO_RAW]
            if False:
                print( 'a_peak_info_raw=', a_peak_info_raw )
                print( 'x_peak_info_raw=', x_peak_info_raw )

            a_num_peaks_raw = len( a_peak_info_raw )
            b_num_peaks_raw = len( x_peak_info_raw )
            if a_num_peaks_raw != b_num_peaks_raw:
                raise RuntimeError( 'Numbers of raw peaks are inconsistent: %d != %d' % ( a_num_peaks_raw, b_num_peaks_raw ),
                                    [ 2, a_num_peaks_raw, b_num_peaks_raw ] )

    def check_peak_concictency( self, a_peak_info, x_peak_info, allowance ):
        if False:
            if len( a_peak_info ) == len( x_peak_info ):
                # better check?
                return [], []

        indeces, params = get_probably_corresponding_index( self.a_curve, self.x_curve )
        if params is None:
            return [], []

        print( 'indeces=', indeces )
        a_index, x_index = indeces

        def get_peaks_to_remove( index1, index2 ):
            # print( 'index1=', index1, 'index2=', index2 )
            to_remove = []
            for i in index1:
                if i not in index2:
                    to_remove.append( i )
            return to_remove

        a_peaks = np.array( [ info[1] for info in a_peak_info ] )
        x_peaks = np.array( [ info[1] for info in x_peak_info ] )

        e_remove = get_peaks_to_remove( list(range(len(a_peaks))), a_index )
        x_remove = get_peaks_to_remove( list(range(len(x_peaks))), x_index )

        print( 'e_remove, x_remove=', e_remove, x_remove )

        return e_remove, x_remove


    def compute_xray_baseline(self, opt_params):
        baseline_type = opt_params['xray_baseline_type']
        if baseline_type == 0:
            return np.zeros(len(self.x_vector))

        y   = copy.deepcopy( self.x_vector )
        with_bpa = opt_params['xray_baseline_with_bpa']
        if baseline_type > 0:
            # since we don't have created the self.x_curve yet, we will use self.sd_xray_curve instead
            # giving curve=temp_x_curve to improve failed LPM cases such as 20181127
            # TODO: unify creation of ScatteringBaseline objects
            ret_y = compute_baseline_using_LPM_impl( baseline_type, self.num_iterations, self.x_x, y, curve=self.sd_xray_curve,
                                                        logger=self.logger, suppress_log=False )

        if with_bpa:
            from molass_legacy.Baseline.LambertBeer import get_standard_baseline
            if self.baseline_corrected_copy is None:
                sd_copy = self.serial_data.get_copy()
                mapped_info = self.prepare_env_for_plain_LPM()
                sd_copy.apply_baseline_correction(mapped_info, basic_lpm=True)
                self.baseline_corrected_copy = sd_copy
                if self.sd_update_callback is not None:
                    self.sd_update_callback(sd_copy)
            sd_ = self.baseline_corrected_copy

            base_y = get_standard_baseline(sd_, None, logger=self.logger, debug=False)
            if False:
                x = self.x_x
                plt.push()
                fig, ax = plt.subplots()
                ax.plot(x, y, label='data')
                ax.plot(x, ret_y, label='corrected')
                ax.plot(x, ret_y + base_y, ':', label='with BPA')
                ax.legend()
                fig.tight_layout()
                plt.show()
                plt.pop()
            ret_y += base_y

        return ret_y

    def prepare_env_for_plain_LPM(self):
        self.x_curve = self.serial_data.xray_curve
        if self.x_curve_update_callback is not None:
            self.x_curve_update_callback(self.x_curve)
        self.x_base = np.zeros( len(self.x_vector) )
        self.x_base_adjustment = 0
        """
        the above three will be referenced in the following correction
        and updated after this method completes.
        i.e., correction is done without any adjustment
        """
        affine_info = self.get_affine_info()
        opt_params = {  'xray_baseline_opt':1,
                        'xray_baseline_type':1,
                        'xray_baseline_adjust':0,
                        'xray_baseline_with_bpa':1}

        # return Struct(opt_params=opt_params, affine_info=affine_info, x_curve=self.x_curve)
        return Struct(opt_params=opt_params, affine_info=affine_info)

    def compute_initial_params(self, debug=False):
        A, B = self.peak_mapper.best_info[1]    # best_params
        self.A_init = A
        self.B_init = B
        self.logger.info( 'approximate mapping params: A=%g, B=%g' % (A, B) )
        self.a_peak_ys  = [ self.a_curve.spline( info[1] ) for info in self.a_curve.peak_info ]
        self.x_peak_ys  = [ self.x_curve.spline( info[1] ) for info in self.x_curve.peak_info ]
        self.map_params = A, B
        self.init_scales = [yx/ya for ya, yx in zip(self.a_peak_ys, self.x_peak_ys)]
        if debug:
            self.logger.info("init_scales=%s", str(self.init_scales))

        if debug:
            from molass_legacy.Elution.CurveUtils import simple_plot

            A_ = 1/A
            B_ = -B/A

            with plt.Dp():
                fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
                ax1, ax2 = axes

                fig.suptitle("compute_initial_params debug")
                simple_plot(ax1, self.a_curve, "UV", color='blue', legend=False)
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)
                simple_plot(ax2, self.x_curve, "Xray", color='orange', legend=False)
                mapped_tops = []
                for top in self.a_curve.peak_top_x:
                    mapped_tops.append(A_*top + B_)

                ax2.plot(mapped_tops, self.x_curve.spline(mapped_tops), 'o', color='pink', label='mapped_tops')

                ax1.legend()
                ax2.legend()
                fig.tight_layout()
                fig.subplots_adjust(top=0.9)
                plt.show()

    def check_mapping_normality( self, A, B, std_diff=None ):
        """
        note that this method is usually called twice, before and after optimazation
        """
        x_ends  = self.x_x[ [0, -1] ]
        a_ends  = self.a_x[ [0, -1] ]
        mapped_x_ends   = A * x_ends + B
        mapped_x_ratio  = ( mapped_x_ends[1] - mapped_x_ends[0] ) / ( self.a_x[-1] - self.a_x[0] )
        mapped_a_ends   = ( a_ends - B ) / A
        mapped_a_ratio  = ( mapped_a_ends[1] - mapped_a_ends[0] ) / ( self.x_x[-1] - self.x_x[0] )
        # print( '---- mapped_x_ratio=', mapped_x_ratio, 'mapped_a_ratio=', mapped_a_ratio )
        if abs( mapped_x_ratio - 1 ) > MAPPED_RATIO_ALLOW or abs( mapped_a_ratio - 1 ) > MAPPED_RATIO_ALLOW:
            if True:
                self.logger.warning('abnormal mapping suspected with mapped_x_ratio=%.3g and mapped_a_ratio=%.3g' % (mapped_x_ratio, mapped_a_ratio))
            else:
                return RuntimeError( "Approximate mapping is not normal: mapped_x_ratio=%.3g, mapped_a_ratio=%.3g"
                                    % ( mapped_x_ratio, mapped_a_ratio ),
                                    [ 3, mapped_x_ratio, mapped_a_ratio, mapped_x_ends, mapped_a_ends ] )
        return None

    def compute_approximate_std_diff( self, A, B ):
        x_peak  = self.x_curve.peak_info[0][1]
        x_pk_y  = self.x_curve.spline(x_peak)
        x_y     = self.x_curve.y
        size    = len(x_y)

        a_peak  = self.a_curve.peak_info[0][1]
        a_pk_y  = self.a_curve.spline(a_peak)
        a_x     = A * self.x_x + B
        a_y     = self.a_curve.spline( a_x )
        a_y_scaled = a_y * x_pk_y / a_pk_y

        if False:
            from CanvasDialog   import CanvasDialog
            def plot_func( fig, parent ):
                ax  = fig.add_subplot( 111 )
                ax.plot( a_y_scaled )
                ax.plot( x_y, color='orange' )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug: a_y_scaled", adjust_geometry=True )
            dialog.show( plot_func, figsize=(16, 8), parent_arg=True )

        std_chi_square  = np.sum( ( x_y - a_y_scaled )**2 ) / size
        std_diff        = np.sqrt( std_chi_square ) / x_pk_y * STD_DIFF_SCALE
        # print( 'approximate_std_diff=', std_diff )
        return std_diff, a_y_scaled, x_y

    def optimize_mapping(self, devel=False):
        if devel:
            from importlib import reload
            import Mapping.MappingOptimizer
            reload(Mapping.MappingOptimizer)
        from .MappingOptimizer import optimize_mapping_impl

        optimize_mapping_impl(self)

    def determine_mapping_ranges( self ):
        self.x_ranges = self.x_curve.get_ranges_by_ratio( MAPPING_RANGE_RATIO, debug=False )
        # self.mapping_ranges = self.x_curve.peak_info      # there exists a bug for 20181203
        self.mapping_ranges = self.x_ranges             # temp fix for the above bug; must be verified for other cases

        if REFINE_MAPPING_RANGES:
            try:
                self.refine_mapping_ranges()
            except:
                etb = ExceptionTracebacker()
                self.logger.warning("refine_mapping_ranges faild: %s", str(etb))

    def refine_mapping_ranges( self ):
        # not yet usable for such cases 20180206
        # print('before refine', self.mapping_ranges)

        x_ranges = self.x_ranges
        x_mapping_ranges = self.mapping_ranges

        a_mapping_ranges = self.a_curve.peak_info
        x_delta = int( len( self.x_curve.x ) * MAPPING_RANGE_DELTA + 0.5 )

        for i, a_boundary in enumerate( self.a_curve.boundaries ):
        # for i, a_boundary in enumerate( a_boundaries ):
            x_boundary  = self.x_curve.boundaries[i]
            xm_boundary = self.A_init * x_boundary + self.B_init
            x_lower     = x_mapping_ranges[i][2]
            x_upper     = x_mapping_ranges[i+1][0]
            a_lower     = a_mapping_ranges[i][2]
            a_upper     = a_mapping_ranges[i+1][0]
            am_boundary = int( ( a_boundary - self.B_init )/self.A_init + 0.5 )
            # print( 'refine_mapping_ranges (1)', [i], a_lower, a_boundary, xm_boundary, a_upper )

            if False:
                fig = plt.figure( figsize=(12,6) )
                ax1 = fig.add_subplot( 121 )
                ax2 = fig.add_subplot( 122 )

                b_colors = [ 'green', 'yellow' ]

                ax1.plot( self.a_curve.y )
                ymin1, ymax1 = ax1.get_ylim()
                ax1.set_ylim( ymin1, ymax1 )
                for x in [a_lower, a_upper]:
                    ax1.plot( [ x, x ], [ ymin1, ymax1 ], ':', color='black', alpha=0.2 )
                for k, x in enumerate([a_boundary, xm_boundary]):
                    ax1.plot( x, self.a_curve.spline(x), 'o', color=b_colors[k] )

                ax2.plot( self.x_curve.y, color='orange' )
                ymin2, ymax2 = ax2.get_ylim()
                ax2.set_ylim( ymin2, ymax2 )
                for x in [x_lower, x_upper]:
                    ax2.plot( [ x, x ], [ ymin2, ymax2 ], ':', color='black', alpha=0.2 )
                for k, x in enumerate([am_boundary, x_boundary]):
                    ax2.plot( x, self.x_curve.spline(x), 'o', color=b_colors[k] )

                fig.tight_layout()
                plt.show()

            top_x = x_mapping_ranges[i][1]
            if am_boundary < x_lower:
                am_boundary_    = am_boundary - x_delta
                upper_candidate = min( x_ranges[i][2], am_boundary_ )
                if upper_candidate > top_x:
                    x_mapping_ranges[i][2] = am_boundary_
                    x_ranges[i][2] = upper_candidate
                else:
                    self.logger.warning('failed to refine right boundary of %s to %d < %d' % ( str(x_mapping_ranges[i]), upper_candidate, top_x) )

            if am_boundary > x_upper:
                am_boundary_    = am_boundary + x_delta
                lower_candidate = max( x_ranges[i+1][0], am_boundary_ )
                if lower_candidate < top_x:
                    x_mapping_ranges[i+1][0] = am_boundary_
                    x_ranges[i+1][0] = lower_candidate
                    self.logger.warning('failed to refine ')
                else:
                    self.logger.warning('failed to refine left boundary of %s to %d > %d' % ( str(x_mapping_ranges[i]), lower_candidate, top_x) )

        self.x_ranges   = x_ranges
        self.mapping_ranges = x_mapping_ranges
        # print('after refine', self.mapping_ranges)

    def make_whole_mapped_vector(self, A, B, scales, make_inv_mapped_boundaries=False, with_original_scale=False ):
        boundaries = self.x_curve.boundaries
        # print( 'mapping_ranges=', self.mapping_ranges )
        # print( 'boundaries=', boundaries )

        debug = False
        detail_debug = False

        mapped_vector_list  = []
        if detail_debug:
            debug_list = []
        if with_original_scale:
            orig_vector_list = []

        for i, range_ in enumerate( self.mapping_ranges ):
            peak    = self.x_curve.peak_info[i][1]
            p_start = max(0, range_[0])     # temp fix to avoid range_[0] < 0 for raw Kosugi8, which should be investigated
            p_stop  = min( len(self.x_curve.y), range_[2] + 1 )
            xray_x  = np.arange( p_start, p_stop )
            S = scales[i]

            if i == 0:
                if debug: print( 'mapped vector (1)', ( 0, p_start ) )
                x_  = np.arange( 0, p_start )
                m_vec, orig_vec = self.make_mapped_vector( x_, A, B, S )
                mapped_vector_list.append( m_vec )
                if detail_debug:
                    debug_list.append([0, (0, p_start), p_start - 0, len(m_vec)])
                if with_original_scale:
                    orig_vector_list.append( orig_vec )

            else:
                if i <= len(boundaries):
                    boundary = int(boundaries[i-1])
                    if debug: print( 'mapped vector (2)', ( prev_stop, p_start ) )
                    m_vec, orig_vec   = self.make_mapped_valley_vector( i, prev_stop, boundary, p_start, prev_A, prev_B, prev_S, A, B, S, make_inv_mapped_boundaries,
                                            with_original_scale=with_original_scale)
                    mapped_vector_list.append( m_vec )
                    if detail_debug:
                        debug_list.append([2, ( prev_stop, p_start ), p_start - prev_stop, len(m_vec)])

                    if with_original_scale:
                        orig_vector_list.append( orig_vec )

            if debug: print( 'mapped vector (3)', ( p_start, p_stop ) )
            prev_A, prev_B, prev_S, prev_stop = A, B, S, p_stop
            m_vec, orig_vec = self.make_mapped_vector( xray_x, A, B, S )
            mapped_vector_list.append( m_vec )
            if detail_debug:
                debug_list.append([3, ( p_start, p_stop), p_stop - p_start, len(m_vec)])

            if with_original_scale:
                orig_vector_list.append( orig_vec )

            if i == len( self.mapping_ranges ) - 1:
                if debug: print( 'mapped vector (4)', ( p_stop, len(self.x_vector) ) )
                x_  = np.arange( p_stop, len(self.x_vector) )
                m_vec, orig_vec = self.make_mapped_vector( x_, A, B, S )
                mapped_vector_list.append( m_vec )
                if detail_debug:
                    debug_list.append([4, (p_stop, len(self.x_vector)), len(self.x_vector) - p_stop, len(m_vec)])
                if with_original_scale:
                    orig_vector_list.append( orig_vec )

            if debug:
                temp_mvec = np.hstack( mapped_vector_list )
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title( "%d-th plot with range_=%s" % (i, str(range_)) )
                    ax.plot( self.x_curve.x, self.x_curve.y )
                    ax.plot( temp_mvec )
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    x_  = len(temp_mvec)
                    ax.plot( [x_, x_], [ymin, ymax], ':', color='red' )
                    plt.tight_layout()
                    plt.show()

        if debug:
            length = 0
            for k, vec in enumerate(mapped_vector_list):
                length += len(vec)
                print([k], len(vec), length)
            print(len(self.x_x))

        mapped_vector  = np.hstack( mapped_vector_list )
        if detail_debug and len(mapped_vector) != len(self.x_curve.x):
            print('len(mapped_vector)=', len(mapped_vector))
            print('len(self.x_curve.x)=', len(self.x_curve.x))
            print('self.mapping_ranges=', self.mapping_ranges)
            print('boundaries=', boundaries)

            x = self.x_curve.x
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.plot(x, self.x_curve.y)
                ax.plot(mapped_vector)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                mv_start = 0
                for k, mv in enumerate(mapped_vector_list):
                    info = debug_list[k]
                    print( [k], mv_start, len(mv), info )
                    if len(mv) > 0:
                        f = mv_start
                        t = mv_start + len(mv) - 1
                        for p in [f, t]:
                            ax.plot([p, p], [ymin, ymax], ':', color='yellow')
                        mv_start += len(mv)

                plt.tight_layout()
                plt.show()

        if with_original_scale:
            orig_vector = np.hstack( orig_vector_list )
        else:
            orig_vector = None

        assert len(mapped_vector) == len(self.x_x)
        return mapped_vector, orig_vector

    def make_mapped_vector( self, xray_x, A, B, S ):
        j_  = A*xray_x + B
        if False:
            ax = plt.gca()
            ax.set_title('make_mapped_vector debug')
            try:
                ax.plot(self.a_curve_spline_adjusted( j_ ), label='a_curve_spline_adjusted')
            except:
                pass
            ax.plot(self.a_curve.spline( j_ ), label='a_curve.spline')
            ax.legend()
            plt.tight_layout()
            plt.show()

        if self.in_uv_adjustment_mode:
            tgt_y   = self.a_curve_spline_adjusted( j_ )
        else:
            tgt_y   = self.a_curve.spline( j_ )
        m_vec   = tgt_y * S
        return m_vec, tgt_y

    def make_mapped_valley_vector( self, i, prev_stop, x_boundary, p_start, prev_A, prev_B, prev_S, A, B, S, make_inv_mapped_boundaries,
                                     with_original_scale=False):
        # print( 'make_mapped_valley_vector: prev_stop, x_boundary, p_start=', prev_stop, x_boundary, p_start )

        if x_boundary < prev_stop:
            # temporary fix for 20181203 when it is not an analysis copy
            self.logger.warning("boundary has been modified from %d to %d in make_mapped_valley_vector as a temporary fix." % (x_boundary, prev_stop))
            x_boundary = prev_stop
        if x_boundary > p_start:
            # temporary fix for 20171203
            self.logger.warning("boundary has been modified from %d to %d in make_mapped_valley_vector as a temporary fix." % (x_boundary, p_start))
            x_boundary = p_start

        xray_x1 = np.arange( prev_stop, x_boundary )
        xray_x2 = np.arange( x_boundary, p_start )

        if False:
            print( 'prev_stop, x_boundary, p_start=', prev_stop, x_boundary, p_start )
            print( 'xray_x1=', len(xray_x1), xray_x1 )
            print( 'xray_x2=', len(xray_x2), xray_x2 )
            x_curve = self.x_curve
            x   = x_curve.x
            y   = x_curve.y
            spline  = x_curve.spline
            fig = plt.figure()
            ax = fig.gca()
            ax.set_title("make_mapped_valley_vector debug")
            ax.plot( x, y )
            for x_ in [prev_stop, x_boundary, p_start]:
                ax.plot( x_, spline(x_), 'o' )
            plt.show()

        denom_width = p_start - prev_stop
        if denom_width > 0:
            w0 = ( x_boundary - 1 - prev_stop ) / denom_width
        else:
            # as occurred in 20191006_proteins5
            w0 = 0

        if USE_SAME_RATIO_BOUNDARY:
            # determine a_boundary so that a_boundary keeps the same ratio as x_boundary
            # in relation to ( prev_stop, p_start  ) and ( v_lower, v_upper ) respectively
            v_lower = prev_A * prev_stop + prev_B
            v_upper = A      * p_start   + B
            a_boundary  = v_lower * ( 1 - w0 ) + v_upper * w0

            # determine A_ so that it is consistent with the interval length ratio
            A_  = ( v_upper - v_lower ) / ( p_start - prev_stop )

            # determine B_ so that a_boundary be mapped to x_boundary
            B_  = a_boundary - A_ * x_boundary
        else:
            A_  = prev_A * ( 1 - w0 ) + A * w0
            B_  = prev_B * ( 1 - w0 ) + B * w0
            a_boundary  = A_ * x_boundary + B_

        S_      = prev_S * ( 1 - w0 ) + S * w0
        S_min   = min( prev_S, S )
        S_max   = max( prev_S, S )

        if make_inv_mapped_boundaries:
            self.inv_mapped_boundaries.append( a_boundary )

        if len(xray_x1) > 1:
            w1  = np.arange( len(xray_x1) ) / ( len(xray_x1) - 1)

            last_j  = prev_A*xray_x1 + prev_B
            this_j  = A_*xray_x1 + B_
            j_      = last_j * ( 1 - w1 ) + this_j * w1

            if self.in_uv_adjustment_mode:
                tgt_y1  = self.a_curve_spline_adjusted( j_ )
            else:
                tgt_y1  = self.a_curve.spline( j_ )

            if USE_SIMPLE_SCALE_FOR_VALLEY:
                # this scaling seems better as in 20170209/OA_Ald_Fer
                boundary_scale = S_
            else:
                a_by    = tgt_y1[-1]
                xj      = xray_x1[-1]
                if self.in_xray_adjustment_mode:
                    x_by    = self.x_curve_y_adjusted[ xj ]
                else:
                    x_by    = self.x_curve.y[ xj ]
                if abs( a_by ) > SMALL_VALUE_LIMIT:
                    b_scale = x_by / a_by
                    # avoid extreme values
                    abs_b_scale = max( S_min/2, min( S_max*2, abs( b_scale )  ) )
                    boundary_scale  = abs_b_scale if b_scale >= 0 else -abs_b_scale
                else:
                    boundary_scale  = S_

            # print( 'scales=', prev_S, boundary_scale, S )
            scale1  = prev_S * ( 1 - w1 ) + boundary_scale * w1
            m_vec1  = tgt_y1 * scale1
        elif len(xray_x1) == 1:
            # 20181204 
            boundary_scale = S_
            j_ = A_*xray_x1 + B_
            tgt_y1  = self.a_curve.spline( j_ )
            m_vec1  = tgt_y1 * S_
        else:
            boundary_scale = S_
            tgt_y1 = np.array([])
            m_vec1 = np.array([])

        if len(xray_x2) > 1:
            w2  = np.arange( len(xray_x2) ) / ( len(xray_x2) - 1)

            last_j  = A_*xray_x2 + B_
            this_j  = A *xray_x2 + B
            j_      = last_j * ( 1 - w2 ) + this_j * w2

            if self.in_uv_adjustment_mode:
                tgt_y2   = self.a_curve_spline_adjusted( j_ )
            else:
                tgt_y2   = self.a_curve.spline( j_ )

            scale2  = boundary_scale * ( 1 - w2 ) + S * w2
            m_vec2  = tgt_y2 * scale2
        elif len(xray_x2) == 1:
            j_ = A_*xray_x2 + B_
            tgt_y2  = self.a_curve.spline( j_ )
            m_vec2  = tgt_y2 * S
        else:
            tgt_y2 = np.array([])
            m_vec2 = np.array([])

        m_vec = np.hstack( [ m_vec1, m_vec2 ] )
        if with_original_scale:
            orig_vec = np.hstack( [ tgt_y1, tgt_y2 ] )
        else:
            orig_vec = None

        return m_vec, orig_vec

    def adjust_baselines( self, opt_params ):
        uv_baseline_adjust      = opt_params[ 'uv_baseline_adjust' ]
        xray_baseline_adjust    = opt_params[ 'xray_baseline_adjust' ]
        self.dev_allow_ratio    = opt_params[ 'dev_allow_ratio' ]
        self.a_base_adjustment  = 0
        self.x_base_adjustment  = 0
        need_re_optimization    = False

        if USE_AREA_SCALING:
            if uv_baseline_adjust == 1 and xray_baseline_adjust == 1:
                self.compute_deviation_scales()

        # iterate to get better precision for revised self.mapped_vector
        for k in range( AJUSTMENT_ITERATION ):
            adjust_both = True
            if uv_baseline_adjust == 1 and xray_baseline_adjust == 0 or ( 1 - self.dev_allow_ratio ) < VERY_SMALL_RATIO:
                adjust_both = False
                self.adjust_uv_baseline_only()
                need_re_optimization    = True

            if uv_baseline_adjust == 0 and xray_baseline_adjust == 1 or self.dev_allow_ratio < VERY_SMALL_RATIO:
                adjust_both = False
                self.adjust_xray_baseline_only()
                need_re_optimization    = True

            if uv_baseline_adjust == 1 and xray_baseline_adjust == 1 and adjust_both:
                self.adjust_both_baselines( opt_params )
                need_re_optimization    = True

            if need_re_optimization:
                self.optimize_mapping()

    def compute_deviation_scales( self ):
        # compute UV-absorbance area
        A, B, C = self.absorbance.get_baseplane_params()
        end_x   = np.array( [0, len(self.a_vector)-1] )
        floor_end_y = A * self.absorbance.std_wvlen + B * end_x + C
        sbl = ScatteringBaseline( -self.a_vector )
        D, E  = sbl.solve()
        ceil_end_y  = -D * end_x - E
        self.a_area = self.compute_area( end_x, floor_end_y, ceil_end_y )
        if False:
            from CanvasDialog   import CanvasDialog
            def debug_plot( fig ):
                ax  = fig.add_subplot( 111 )
                ax.plot( self.a_vector )
                ax.plot( end_x, floor_end_y )
                ax.plot( end_x, ceil_end_y )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug: a_area", adjust_geometry=True )
            dialog.show( debug_plot, figsize=(8, 8) )

        # compute Xray-scattering area
        # giving curve=self.x_curve to improve failed LPM cases such as 20181127
        sbl = ScatteringBaseline( self.x_vector, curve=self.x_curve )
        A, B  = sbl.solve()
        end_x   = np.array( [0, len(self.x_vector)-1] )
        floor_end_y = A * end_x + B
        sbl = ScatteringBaseline( -self.x_vector )
        D, E  = sbl.solve()
        ceil_end_y  = -D * end_x - E
        self.x_area = self.compute_area( end_x, floor_end_y, ceil_end_y )
        if False:
            from CanvasDialog   import CanvasDialog
            def debug_plot( fig ):
                ax  = fig.add_subplot( 111 )
                ax.plot( self.x_vector )
                ax.plot( end_x, floor_end_y )
                ax.plot( end_x, ceil_end_y )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug: x_area", adjust_geometry=True )
            dialog.show( debug_plot, figsize=(8, 8) )

    def compute_area( self, end_x, floor_end_y, ceil_end_y ):
        return end_x[-1] * ( ceil_end_y[0] - floor_end_y[0] + ceil_end_y[1] - floor_end_y[1]  ) / 2

    def get_adjusted_a_spline(self):
        # a_curve_y = self.a_curve.y
        a_curve_y = self.a_curve.spline(self.a_x)
        ac_vector = a_curve_y - self.a_base_adjustment
        return UnivariateSpline( self.a_x, ac_vector, s=0, ext=3 )

    def uv_baseline_adjuster( self, params ):
        D, E = params
        self.a_base_adjustment = D * self.a_x + E
        self.a_curve_spline_adjusted = self.get_adjusted_a_spline()
        adj_scales = self.adjust_uv_opt_results( D, E )
        mapped_vector, _ = self.make_whole_mapped_vector(*self.map_params, adj_scales)
        diff_y      = self.x_curve.y - mapped_vector

        if False:
            from CanvasDialog   import CanvasDialog
            from molass_legacy.SerialAnalyzer.ElutionCurve   import proof_plot
            print( 'D, E=', D.value, E.value, 'diff2_sum=', np.sum( diff_y**2 ) )
            def plot_func( fig ):
                ax1 = fig.add_subplot( 131 )
                ax2 = fig.add_subplot( 132 )
                ax3 = fig.add_subplot( 133 )
                ax1.plot( self.a_curve.y )
                ax1.plot( self.a_base_adjustment, color='yellow')
                # ax2.plot( ac_vector )
                ax2.plot( self.a_curve_spline_adjusted( self.a_x ), ':' )
                ax3.plot( self.x_curve.y, color='orange' )
                ax3.plot( mapped_vector )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug: a_curve", adjust_geometry=True )
            dialog.show( plot_func, figsize=(24, 6) )

        return np.sum(diff_y**2)

    def adjust_uv_opt_results( self, D, E ):
        adj_results = []
        A, B = self.map_params
        for i, info in enumerate(self.x_curve.peak_info):
            x_peak  = info[1]
            x_py    = self.x_curve.spline( x_peak )
            a_peak  = A * x_peak + B
            a_py    = self.a_curve_spline_adjusted( a_peak )
            S_  = x_py / a_py
            adj_results.append(S_ )
        return adj_results

    def adjust_uv_baseline_only( self ):
        self.in_uv_adjustment_mode  = True
        self.in_xray_adjustment_mode = False

        E_init = np.percentile( self.a_curve.y, 5 )

        init_params = (0, E_init)
        bounds = ((-1e-4, 1e-4), (-0.1, 0.1))

        # result  = minimize( self.uv_baseline_adjuster, params, args=() )
        result  = optimize.minimize(self.uv_baseline_adjuster, init_params, bounds=bounds)

        D, E = result.x

        self.a_base_adjustment  =  D * self.a_x + E
        self.a_curve_spline_adjusted = self.get_adjusted_a_spline()

    def xray_baseline_adjuster( self, params ):
        F, G = params

        x_base_adjustment   = F * self.x_x + G

        adj_scales = self.adjust_xray_opt_results( F, G )
        mapped_vector, _ = self.make_whole_mapped_vector(*self.map_params, adj_scales)
        xc_vector = self.x_curve.y - x_base_adjustment
        diff_y  = xc_vector - mapped_vector
        return np.sum(diff_y**2)

    def adjust_xray_opt_results( self, F, G ):
        adj_results = []

        A, B = self.map_params
        for i, info in enumerate(self.x_curve.peak_info):
            x_peak  = info[1]
            x_py    = self.x_curve.spline( x_peak ) - ( F*x_peak + G )
            a_peak  = A * x_peak + B
            a_py    = self.a_curve.spline( a_peak )
            S_  = x_py / a_py
            adj_results.append(S_)
        return adj_results

    def adjust_xray_baseline_only( self, G_init=0 ):
        self.in_uv_adjustment_mode  = False
        self.in_xray_adjustment_mode = True

        init_params = (0, G_init)
        bounds = ((-1e-4, 1e-4), (G_init-0.1, G_init+0.1))

        # result  = minimize( self.xray_baseline_adjuster, params, args=() )
        result = optimize.minimize(self.xray_baseline_adjuster, init_params, bounds=bounds)

        F, G = result.x
        self.x_base_adjustment  = F * self.x_x + G
        self.x_curve_y_adjusted = self.x_curve.y - self.x_base_adjustment
        self.x_peak_ys_adjusted = [ self.x_curve.spline( info[1] ) - ( F * info[1] + G ) for info in self.x_curve.peak_info ]

    def both_baseline_adjuster( self, params ):
        D, E, F, G = params

        xray_y  = self.x_curve.y - ( F * self.x_x + G )

        a_base_adjustment   = D * self.a_x + E
        x_base_adjustment   = F * self.x_x + G

        self.a_base_adjustment = a_base_adjustment
        self.a_curve_spline_adjusted = self.get_adjusted_a_spline()
        self.x_curve_y_adjusted = self.x_curve.y - x_base_adjustment
        adj_scales = self.adjust_both_opt_results( D, E, F, G )
        mapped_vector, _    = self.make_whole_mapped_vector(*self.map_params, adj_scales)
        mapped_adjustment   = self.make_mapped_adjustment(*self.map_param, D, E )
        penalty_vector  = self.compute_penalty_vector( a_base_adjustment, mapped_adjustment, x_base_adjustment )

        if self.adjust_both_debug:
            from CanvasDialog   import CanvasDialog
            def plot_func( fig ):
                ax = fig.add_subplot( 111 )
                ax.plot( mapped_adjustment )
                ax.plot( x_base_adjustment )
                ax.plot( penalty_vector )
                fig.tight_layout()
            dialog = CanvasDialog( "Debug: mapped_adjustment", adjust_geometry=True )
            dialog.show( plot_func, figsize=(8, 6) )
            if not dialog.applied:
                self.adjust_both_debug = False

        diff_y  = np.abs( xray_y - mapped_vector ) + penalty_vector
        return np.sum(diff_y**2)

    def adjust_both_opt_results( self, D, E, F, G ):
        adj_results = []
        A, B = self.map_params
        for i, info in enumerate(self.x_curve.peak_info):
            x_peak  = info[1]
            x_py    = self.x_curve.spline( x_peak ) - ( F*x_peak + G )
            a_peak  = A * x_peak + B
            a_py    = self.a_curve_spline_adjusted( a_peak )
            S_  = x_py / a_py
            adj_results.append(S_)
        return adj_results

    def compute_penalty_vector( self, a_adj, m_adj, x_adj ):
        """
        the purpose of introducing a penalty vector is to prevent the optimizer
        from getting lost into useless cases where both adjustments grow large
        in the same direction.
        """
        dir_penalty  = np.zeros( len(m_adj) )
        both_positive   = np.logical_and( m_adj > 0, x_adj > 0 )
        both_negative   = np.logical_and( m_adj < 0, x_adj < 0 )
        dir_penalty[both_positive]   = np.min( [ m_adj[both_positive], x_adj[both_positive] ], axis=0 )
        dir_penalty[both_negative]   = -np.max( [ m_adj[both_negative], x_adj[both_negative] ], axis=0 )

        if USE_AREA_SCALING:
            uv_area_ratio   = np.sum( np.abs(a_adj) ) / self.a_area
            xray_area_ratio = np.sum( np.abs(x_adj) ) / self.x_area
        else:
            uv_area_ratio   = np.sum( np.abs(m_adj) )
            xray_area_ratio = np.sum( np.abs(x_adj) )

        total_ratio     = uv_area_ratio + xray_area_ratio
        uv_weight   = 1 - self.dev_allow_ratio
        xray_weight = self.dev_allow_ratio

        if total_ratio > 0:
            weight_dev_penalty = 1 + abs( uv_area_ratio/total_ratio * uv_weight - xray_area_ratio/total_ratio * xray_weight )
        else:
            weight_dev_penalty = 1

        if USE_AREA_SCALING:
            x_area  = self.x_area
        else:
            x_area  = self.x_curve.max_y * ( len( self.x_vector ) - 1 ) / 2
            # * 1/2 is an approximate scalling to make it equivalent to USE_AREA_SCALING

        keep_positive = x_area * PENALTY_KEEP_POS_VALUE

        ret_penalty = ( dir_penalty + keep_positive ) * weight_dev_penalty

        if False:
            from matplotlib.patches import Polygon
            from molass_legacy.KekLib.OurMatplotlib      import get_default_colors
            print( 'x_area=', x_area, 'keep_positive=', keep_positive )
            colors = get_default_colors()

            fig = plt.figure(figsize=(16, 8))
            ax1 = fig.add_subplot( 121 )
            ax2 = fig.add_subplot( 122 )
            ax1.set_title( "UV-absorbance adustment" )
            ax2.set_title( "Xray and mapped adustment with the penalty vector" )
            ax1.plot( a_adj, label='UV-adjustment' )
            ax2.plot( m_adj, label='mapped adjustment' )
            ax2.plot( x_adj, label='Xray adjustment' )
            ax2.plot( dir_penalty, ':', label='same direction penality' )
            ax2.plot( ret_penalty, ':', label='penalty with deviation ratio' )
            pos = np.where( dir_penalty > 0 )[0]
            if len(pos) > 0:
                null = np.array( [], dtype=int )
                pos_x = np.hstack( [ [pos[0]-1] if pos[0] > 0 else null, pos, [pos[-1]+1] if pos[-1] < len(dir_penalty)-1 else null ] )
                pos_y = dir_penalty[pos_x]
                points = list( zip( pos_x, pos_y ) )
                polygon = Polygon( points, alpha=0.2, fc=colors[2] )
                ax2.add_patch( polygon )
            ax1.legend()
            ax2.legend()
            fig.tight_layout()
            plt.show()

        return ret_penalty

    def make_mapped_adjustment(self, A, B, D, E ):
        adjustment_array = []

        effective_num_boundaries = min(len(self.x_curve.boundaries), len(self.scale_params) - 1)
        if effective_num_boundaries < len(self.x_curve.boundaries):
            """
                this case occurred in 20190221_2 with LPM(integral)
                may better be avoided by computing self.opt_results for each peak
            """
            self.logger.warning("wrong number of opt_results: len(self.opt_results) < len(self.x_curve.boundaries) + 1")

        start = 0
        for k, S in enumerate( self.scale_params ):
            if k < effective_num_boundaries:
                stop    = self.x_curve.boundaries[k]
            else:
                stop    = len(self.x_x)
            i_  = np.arange( start, stop )
            j_  = A * i_ + B
            adjustment = ( D * j_ + E ) * S
            adjustment_array.append( adjustment )
            start   = stop

        mapped_adjustment = np.hstack( adjustment_array )
        assert len(mapped_adjustment) == len(self.x_x)
        return mapped_adjustment

    def adjust_both_baselines( self, opt_params, G_init=0 ):
        self.in_uv_adjustment_mode      = True
        self.in_xray_adjustment_mode    = True

        init_params = (0, 0, 0, G_init)
        bounds = ((-1e-4, 1e-4), (-0.1, 0.1 ), (-1e-4, 1e-4 ), (G_init-0.1, G_init+0.1))

        self.adjust_both_debug = False

        # result  = minimize( self.both_baseline_adjuster, params, args=() )
        result = optimize.minimize(self.both_baseline_adjuster, init_params, bounds=bounds)

        D, E, F, G = result.x
        self.a_base_adjustment  = D * self.a_x + E
        self.a_curve_spline_adjusted = self.get_adjusted_a_spline()
        self.x_base_adjustment  = F * self.x_x + G
        self.x_curve_y_adjusted = self.x_curve.y - self.x_base_adjustment
        self.x_peak_ys_adjusted = [ self.x_curve.spline( info[1] ) - ( F * info[1] + G ) for info in self.x_curve.peak_info ]

    def compute_curve_for_eval_diff( self ):
        pass

    def get_uniformly_scaled_vector( self ):
        if self.uniformly_scaled_vector is None:
            self.uniformly_scaled_vector = self.make_uniformly_scaled_vector()

        return self.uniformly_scaled_vector

    def get_x2a_scale(self):
        x_peak_y_adjust = 0 if np.isscalar( self.x_base_adjustment) else self.x_base_adjustment[ self.x_curve.primary_peak_i ]
        a_j = int( self.A_init * self.x_curve.primary_peak_i + self.B_init + 0.5 )
        a_peak_y_adjust = 0 if np.isscalar( self.a_base_adjustment) else self.a_base_adjustment[ a_j ]

        scale   = ( self.a_curve.spline(a_j)- a_peak_y_adjust ) / ( self.x_curve.primary_peak_y - x_peak_y_adjust )
        return scale

    def make_uniformly_scaled_vector( self, scale=None, survey_fh=None):
        if scale is None:
            scale = self.get_x2a_scale()

        if np.isscalar(self.a_base_adjustment):
            adj_spline  = lambda x: self.a_base_adjustment
        else:
            adj_spline  = UnivariateSpline( self.a_x, self.a_base_adjustment, s=0, ext=3 )

        if USE_SIMPLE_MAPPING:
            use_simple_mapping = True
        else:
            scaled_vector_array = []
            start = 0
            last_joint_y = None
            use_simple_mapping = False
            A, B = self.map_params
            for k, _ in enumerate( self.x_curve.peak_info ):
                if k < len( self.x_curve.boundaries ):
                    stop    = self.x_curve.boundaries[k]
                else:
                    stop    = len(self.x_curve.x)
                i_  = self.x_curve.x[start:stop]
                j_  = A*i_ + B
                tgt_y   = self.a_curve.spline( j_ ) - adj_spline( j_ )
                if last_joint_y is not None:
                    joint_diff_ratio = abs(tgt_y[0] - last_joint_y)/self.a_curve.max_y
                    if joint_diff_ratio > 0.1:
                        use_simple_mapping = True
                        self.logger.info('a significant gap with ratio %.3g detected in uniformly scaled curve' % joint_diff_ratio)
                        break
                last_joint_y = tgt_y[-1]
                scaled_vector_array.append( tgt_y/scale )
                start   = stop

        if use_simple_mapping:
            scaled_vector_array = self.make_uniformly_scaled_vector_in_another_way(scale, adj_spline, survey_fh=survey_fh)

        return np.hstack( scaled_vector_array )

    def make_uniformly_scaled_vector_in_another_way( self, scale, adj_spline, survey_fh=None):
        # print('make_uniformly_scaled_vector_in_another_way: scale=', scale)
        if True:
            A, B = self.map_params
        else:
            if self.A_simple is None:
                self.compute_simple_params()
            A, B = self.A_simple, self.B_simple
        if survey_fh is not None:
            from .SurveyUtils import write_info
            # this case is temporary only for change influence survey purpose
            write_info(survey_fh, self)

        j_ = A * self.x_curve.x + B
        tgt_y  = ( self.a_curve.spline( j_ ) - adj_spline( j_ ) ) / scale
        # tgt_y2 = self.a_curve_spline_adjusted( j_ ) / scale
        if False:
            fig = plt.figure(figsize=(12,6))
            fig.suptitle("make_uniformly_scaled_vector_in_another_way")
            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)
            ax1.plot( tgt_y, label='tgt_y' )
            ax1.plot( tgt_y, label='tgt_y2' )
            ax2.plot( self.x_curve.y, color='orange' )
            fig.tight_layout()
            plt.show()
        return tgt_y

    def make_scaled_xray_curve_y(self, debug=False):
        scaled_vector_list = []

        x = self.x_curve.x
        y = self.x_curve_y_adjusted

        start = 0
        A, B = self.map_params
        for k, _ in enumerate( self.x_curve.peak_info ):
            if k < len( self.x_curve.boundaries ):
                stop    = self.x_curve.boundaries[k]
            else:
                stop    = len(x)
            S = self.scale_params[k]
            scaled_vector_list.append(y[start:stop]/S)
            start = stop

        scaled_y = np.hstack(scaled_vector_list)

        if debug:
            from molass_legacy.Elution.CurveUtils import simple_plot
            from DataUtils import get_in_folder

            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21, 7))
            ax1, ax2, ax3 = axes
            ax3t = ax3.twinx()

            fig.suptitle("Proof Plot of Mapping for " + get_in_folder(), fontsize=20)
            x_ = self.inv_mapped_boundaries
            ax1.plot(x_, self.a_curve.spline(x_), 'o', color='yellow', markersize=10, label='invserse mapped')
            simple_plot(ax1, self.a_curve)
            simple_plot(ax2, self.x_curve)
            simple_plot(ax3, self.x_curve)
            ax3t.plot(x, scaled_y, ':', label='scaled_y', color='green')

            ax3t.legend()
            fig.tight_layout()
            fig.subplots_adjust(top=0.9)

            plt.show()

        return scaled_y

    def get_mapped_info( self ):
        affine_info = self.get_affine_info()
        return MappedInfo( opt_results=self.opt_results, a_curve=self.a_curve, x_curve=self.x_curve, x_ranges=self.x_ranges,
                            affine_info=affine_info, flow_changes=self.flow_changes,
                            opt_params=self.opt_params, x_base=self.x_base )

    def get_affine_info( self ):
        peaks   = [ info[1] for info in self.x_curve.peak_info ]
        peak_x  = np.average( peaks )
        peak_y  = np.average( self.x_curve.spline( peaks ) )
        return [ self.x_vector, self.x_base, self.x_base_adjustment, (peak_x, peak_y), peaks ]

    def get_num_peaks( self ):
        return len(self.a_curve.peak_info), len(self.x_curve.peak_info)

    def get_conc_vector( self, conc_factor ):
        conc_curve_type = get_setting('conc_curve_type')
        if conc_curve_type == 0:
            ret_uv_vector = self.make_uniformly_scaled_vector(scale=1)
            curve_source = 'Mapped UV Elution'
        else:
            ret_uv_vector = self.make_scaled_xray_curve_y()
            curve_source = 'Scaled Xray Elution'
        self.logger.info("concentration curve made from " + curve_source)
        return ret_uv_vector * conc_factor

    def set_peak_eval_ranges( self ):
        ranges_ = self.x_curve.get_ranges_by_ratio(PEAK_EVAL_RANGE_RATIO)

        self.peak_eval_ranges = []
        for range_ in ranges_:
            p = range_[1]
            for mrange in self.mapping_ranges:
                if abs( p - mrange[1] ) < 5:
                    self.peak_eval_ranges.append( range_ )
                    break
 
        self.uv_peak_eval_ranges = []
        A, B = self.map_params
        for k, range_ in enumerate(self.peak_eval_ranges):
            self.uv_peak_eval_ranges.append( [ A*x + B for x in range_ ] )

    def compute_std_diff( self ):
        if self.in_xray_adjustment_mode:
            x_curve_y   = self.x_curve_y_adjusted
        else:
            x_curve_y   = self.x_curve.y

        if USE_TOTAL_STD_DIFF:
            size    = len(self.x_curve.y)
            if USE_WEIGHTED_DIFF:
                weight  = np.ones( size )
                for lower, peak_i, upper in self.peak_eval_ranges:
                    weight[lower:upper+1]   = VALLEY_DIFF_WEIGHT
            else:
                weight = 1
            std_chi_square = np.sum( weight * ( self.mapped_vector - x_curve_y )**2 ) / size
            std_diff = np.sqrt( std_chi_square ) / x_curve_y[self.x_curve.primary_peak_i]
        else:
            std_diff = 0
            size    = 0
            for lower, peak_i, upper in self.peak_eval_ranges:
                size    = upper - lower + 1
                slice_  = slice( lower, upper+1 )
                diff_y  = self.mapped_vector[slice_] - x_curve_y[slice_]
                std_chi_square  = np.sum( diff_y**2 ) / size
                std_diff += np.sqrt( std_chi_square ) / x_curve_y[peak_i]

        self.std_diff   = std_diff * STD_DIFF_SCALE

    def get_sci_list( self ):
        self.sci_list = self.scomp.compute_sci()
        return self.sci_list

    def get_min_sci(self):
        sci_array = np.array(self.get_sci_list())
        return np.min(sci_array)

    def get_opt_params_from_xe( self, xe ):
        start = 0
        i = None
        for k, b in enumerate( self.x_curve.boundaries ):
            if xe >= start and xe <= b:
                i = k
                break
            start = b
        if i is None:
             i = -1

        return self.opt_results[i]

    def get_uv_value( self, xe ):
        ue  = None
        print( 'mapping_ranges=', self.mapping_ranges )
        print( 'x_curve.boundaries=', self.x_curve.boundaries )

        A, B, S = self.get_opt_params_from_xe( xe )
        ue = A*xe + B

        print( 'get_uv_value: ', xe, ue )
        return ue, self.a_curve.spline(ue), S

    def get_opt_params_from_ue( self, ue ):
        start = 0
        i = None
        assert len(self.a_curve.boundaries) == len(self.x_curve.boundaries)
        for k, b in enumerate( self.a_curve.boundaries ):
            if ue >= start and ue <= b:
                i = k
                break
            start = b
        if i is None:
             i = -1
        return self.opt_results[i]

    def get_mapped_flow_changes( self ):
        fc_xe_list = []
        for fc in self.flow_changes:
            if fc is None:
                fc_xe = None
            else:
                A, B, S = self.get_opt_params_from_ue( fc )
                fc_xe = max(0, min( len(self.x_curve.x)-1, int( ( fc - B )/A + 0.5 ) ) )
            fc_xe_list.append( fc_xe )

        return fc_xe_list

    def get_int_ranges( self ):
        return [ [ int(p+0.5) for p in range_  ]  for range_ in self.mapping_ranges ]

    def get_adjustment_deviation( self, debug=False ):

        if debug:
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12, 6))
            ax1, ax2 = axes
            ax1.plot(self.a_curve.y)
            ax1.plot(self.a_base, color='red')
            ax2.plot(self.x_curve.y, color='orange')
            ax2.plot(self.x_base, color='red')
            fig.tight_layout()
            plt.show()

        a_adj_dev = np.average(np.abs(self.a_base_adjustment))/self.a_curve.max_y
        x_adj_dev = np.average(np.abs(self.x_base_adjustment))/self.x_curve.max_y
        return a_adj_dev + x_adj_dev

    def make_proof_info(self):
        ret_info = []
        for curve in [self.a_curve, self.x_curve]:
            peak_tops = [int(x) for x in curve.peak_top_x]
            boundaries = [int(x) for x in curve.boundaries]
            ret_info.append([peak_tops, boundaries])
        return ret_info

    def compute_simple_params(self):
        if len(self.a_curve.peak_info) <= 1:
            return

        y = self.a_curve.peak_top_x
        x = self.x_curve.peak_top_x

        def objective(p):
            a, b = p
            return np.sum((a*x + b - y)**2)

        ret = optimize.minimize(objective, (self.A_init, self.B_init))

        self.A_simple = ret.x[0]
        self.B_simple = ret.x[1]



"""

    Optimize.ElutionComposer.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
import numpy as np
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting

COMPOSER_CB_TEXT = "try reducing number of components using composite models"

def make_composites_from_deviations(devs):
    allow = get_setting("identification_allowance")

    composites = []
    comp = []
    for j in range(len(devs)):
        comp.append(j)
        if devs[j] < allow:
            pass
        else:
            composites.append(comp)
            comp = []

    comp.append(j+1)
    composites.append(comp)
    return composites

class ElutionComposer:
    def __init__(self, optimizer, params, sd):
        self.logger = logging.getLogger(__name__)
        self.n_components = optimizer.n_components
        self.make_judge_info(optimizer, params, sd)

    def make_judge_info(self, optimizer, params, sd, debug=False):
        from Kratky.GuinierKratkyInfo import GuinierKratkyInfo

        lrf_info = optimizer.objective_func(params, return_lrf_info=True)

        gk_info = GuinierKratkyInfo(optimizer, params, lrf_info)
        devs = gk_info.compute_adjacent_deviation_ratios()

        if debug:
            print("devs=", devs)
            qrgnys = gk_info.qrgnys
            qrgs = gk_info.qrgs
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("Kratky Plot in ElutionComposer debug")

                for k, (qrg, qrgny) in enumerate(zip(qrgs, qrgnys)):
                    ax.plot(qrg, qrgny, 'o', markersize=1, label="component-%d" % (k+1))

                ax.legend()
                fig.tight_layout()
                plt.show()

        self.composites = make_composites_from_deviations(devs) + [[self.n_components - 1]]     # add baseline

    def reducible(self):
        return np.max([len(comp) for comp in self.composites]) > 1

    def make_composite(self):
        from .CompositeInfo import CompositeInfo
        composite = CompositeInfo(composites=self.composites)
        return composite



# coding: utf-8
"""
    InputTeller.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.SharedArrays import SharedArrays

class InputInfo:
    def __init__(self, code, log=True, **kwargs):
        self.code = code
        self.log = log
        self.__dict__.update(kwargs)

    def __repr__(self):
        return 'TellerInfo<%s>' % self.code

class InputTeller:
    def __init__(self, log_folder=None, debug=False):
        import logging
        self.logger = logging.getLogger(__name__)
        self.log_folder = log_folder
        self.debug = debug
        self.invoke_input_manager()
        self.xr_data = None
        self.uv_data = None

    def invoke_input_manager(self):
        import multiprocessing as mp
        from .InputManager import InputManager
        self.logger.info('invoke_input_manager')
        self.input_manager = InputManager(debug=self.debug)
        self.job_queue = mp.Queue()
        self.ret_queue = mp.Queue()
        self.manager_process = mp.Process(target=self.input_manager.manage_loop, args=(self.job_queue, self.ret_queue, self.log_folder))
        self.manager_process.start()
        # print(__name__, 'invoke_input_manager', self.manager_process)

    def __del__(self):
        self.stop()

    def stop(self):
        # print(__name__, 'stop', self.manager_process)
        if self.manager_process is not None:
            print(__name__, 'sending stop')
            self.job_queue.put(InputInfo('stop'))
            self.manager_process.join()
            self.manager_process = None

    def tell(self, code, wait=True, **kwargs):
        self.job_queue.put(InputInfo(code, wait=wait, **kwargs))
        if wait:
            ret = self.ret_queue.get()
        else:
            ret = None
        if self.debug:
            print(code, 'ret=', ret)
        return ret

    def start_load_data(self, in_folder, for_sec=True, xray_only=True):
        self.sh_arrays = None
        self.in_folder = in_folder
        self.tell('start_load_data', args=(in_folder, for_sec, xray_only))

    def is_ready(self):
        return self.tell('is_ready', args=None, log=False)

    def get_sarray(self):
        if self.sh_arrays is None:
            name, tuples = self.tell('get_array_info', args=None)
            sarrays = SharedArrays(name=name, tuples=tuples)
            self.sh_arrays = sarrays.get_arrays()

    def get_redundant_xr_data(self):
        import copy
        self.get_sarray()
        return copy.deepcopy(self.sh_arrays[0])

    def get_xr_data(self):
        from XrayData import XrayData
        self.get_sarray()
        sh_array = self.sh_arrays[0]

        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            from MatrixData import simple_plot_3d
            fig = plt.figure()
            M = sh_array[:,:,1].T
            if False:
                ax = fig.add_subplot(111, projection='3d')
                simple_plot_3d(ax, M)
            else:
                ax = fig.gca()
                ax.plot(M[100, :])
            plt.show()

        self.xr_data = XrayData(None, array=sh_array)
        return self.xr_data

    def get_uv_data(self):
        from molass_legacy.UVData import UvData
        self.get_sarray()
        self.uv_data = UvData(self.in_folder, array=self.sh_arrays[1], vector=self.sh_arrays[2])

        return self.uv_data

    def get_folder_info(self):
        return self.tell('get_folder_info', args=None)



"""
    Solvers.ABC.CustomABC.py

        * custom_distance
        * loop callback

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""

from typing import Callable, List, Tuple, TypeVar, Union
import numpy as np
from datetime import datetime, timedelta

from pyabc.inference.smc import (
    logger, model_output, identity, ABCSMC
)

# from pyabc.model import FunctionModel, Model
from pyabc.random_variables import RV, Distribution
from pyabc.sampler import Sampler
from pyabc.acceptor import (
    Acceptor,
    FunctionAcceptor,
    StochasticAcceptor,
    UniformAcceptor,
)
from pyabc.distance import (
    Distance,
    FunctionDistance,
    PNormDistance,
)
from .CustomDistance import CustomFunctionDistance

from pyabc.epsilon import Epsilon, MedianEpsilon
from pyabc.populationstrategy import ConstantPopulationSize, PopulationStrategy
from pyabc.transition import (
    ModelPerturbationKernel,
    MultivariateNormalTransition,
    Transition,
)
from importlib import reload
import Solvers.ABC.CustomInferenceUtil
reload(Solvers.ABC.CustomInferenceUtil)
from Solvers.ABC.CustomInferenceUtil import (
    create_simulate_function,
)
# from pyabc.platform_factory import DefaultSampler
from .CustomSamplers import DefaultSampler
from pyabc.storage import History
from .CustomModel import CustomObjectiveModel, CustomObjectiveFunction
from pyabc.weighted_statistics import effective_sample_size

class CustomABCSMC(ABCSMC):
    def __init__(
        self,
        models: Union[List[CustomObjectiveModel], CustomObjectiveModel, Callable],
        parameter_priors: Union[List[Distribution], Distribution, Callable],
        distance_function: Union[Distance, Callable] = None,
        population_size: Union[PopulationStrategy, int] = 100,
        summary_statistics: Callable[[model_output], dict] = identity,
        model_prior: RV = None,
        model_perturbation_kernel: ModelPerturbationKernel = None,
        transitions: Union[List[Transition], Transition] = None,
        eps: Epsilon = None,
        sampler: Sampler = None,
        acceptor: Acceptor = None,
        stop_if_only_single_model_alive: bool = False,
        max_nr_recorded_particles: int = np.inf,
        callback: Callable = None,
    ):
        def objective_function(parameter, x0):
            # assert False
            return 0

        self.objective_function = objective_function

        if not isinstance(models, list):
            models = [models]
        models = list(map(lambda x:CustomObjectiveFunction.to_model(self.objective_function, x), models))
        self.models = models

        if not isinstance(parameter_priors, list):
            parameter_priors = [parameter_priors]
        self.parameter_priors = parameter_priors

        # sanity checks
        if len(self.models) != len(self.parameter_priors):
            raise AssertionError(
                "Number models and number parameter priors have to agree."
            )

        assert distance_function is not None
 
        self.distance_function = CustomFunctionDistance.to_distance(
            distance_function,
        )

        self.summary_statistics = summary_statistics

        if model_prior is None:
            model_prior = RV("randint", 0, len(self.models))
        self.model_prior = model_prior

        if model_perturbation_kernel is None:
            model_perturbation_kernel = ModelPerturbationKernel(
                len(self.models), probability_to_stay=0.7
            )
        self.model_perturbation_kernel = model_perturbation_kernel

        if transitions is None:
            transitions = [MultivariateNormalTransition() for _ in self.models]
        if not isinstance(transitions, list):
            transitions = [transitions]
        self.transitions: List[Transition] = transitions

        if eps is None:
            eps = MedianEpsilon(median_multiplier=1)
        self.eps = eps

        if isinstance(population_size, int):
            population_size = ConstantPopulationSize(population_size)
        self.population_size = population_size

        if sampler is None:
            sampler = DefaultSampler()
        self.sampler = sampler

        if acceptor is None:
            acceptor = UniformAcceptor()
        self.acceptor = FunctionAcceptor.to_acceptor(acceptor)

        self.stop_if_only_single_model_alive = stop_if_only_single_model_alive
        self.max_nr_recorded_particles = max_nr_recorded_particles
        self.callback = callback

        # will be set later
        self.x_0 = None
        self.history = None
        self._initial_population = None
        self.minimum_epsilon = None
        self.max_nr_populations = None
        self.min_acceptance_rate = None
        self.max_t = None
        self.max_total_nr_simulations = None
        self.max_walltime = None
        self.min_eps_diff = None

        self.init_walltime = None
        self.analysis_id = None

        self._sanity_check()

    def _create_simulate_function(self, t: int):
        """
        Create a simulation function which performs the sampling of parameters,
        simulation of data and acceptance checking, and which is then passed
        to the sampler.

        Parameters
        ----------
        t: int
            Time index

        Returns
        -------
        simulate_one: callable
            Function that samples parameters, simulates data, and checks
            acceptance.

        .. note::
            For some of the samplers, the sampling function needs to be
            serialized in order to be transported to where the sampling
            happens. Therefore, the returned function should be light, and
            in particular not contain references to the ABCSMC class.
        """
        return create_simulate_function(
            t=t,
            model_probabilities=self.history.get_model_probabilities(t - 1),
            model_perturbation_kernel=self.model_perturbation_kernel,
            transitions=self.transitions,
            model_prior=self.model_prior,
            parameter_priors=self.parameter_priors,
            models=self.models,
            summary_statistics=self.summary_statistics,
            x_0=self.x_0,
            distance_function=self.distance_function,
            eps=self.eps,
            acceptor=self.acceptor,
        )

    def run_generation(
        self,
        t: int,
    ) -> dict:
        """Run a single generation.

        Parameters
        ----------
        t: Generation time index to run for.

        Returns
        -------
        ret:
            Dictionary with entries "successful" indicating whether the
            generation terminated successfully,
            and potentially "acceptance_rate".
        """
        # get epsilon for generation t
        current_eps = self.eps(t)
        if current_eps is None or np.isnan(current_eps):
            raise ValueError(
                f"The epsilon threshold {current_eps} is invalid."
            )
        logger.info(f"t: {t}, eps: {current_eps:.8e}.")

        # create simulate function
        simulate_one = self._create_simulate_function(t)

        # population size and maximum number of evaluations
        pop_size = self.population_size(t)
        max_eval = (
            np.inf
            if self.min_acceptance_rate == 0.0
            else pop_size / self.min_acceptance_rate
        )

        # perform the sampling
        logger.debug(f"Submitting population {t}.")
        sample = self.sampler.sample_until_n_accepted(
            n=pop_size,
            simulate_one=simulate_one,
            t=t,
            max_eval=max_eval,
            ana_vars=self._vars(t=t),
        )

        # check sample health
        if not sample.ok:
            logger.info("Stopping: sample not ok.")
            return {
                "successful": False,
            }

        # normalize accepted population weight to 1
        sample.normalize_weights()

        # retrieve accepted population
        population = sample.get_accepted_population()
        logger.debug(f"Population {t} done.")

        # save to database
        n_sim = self.sampler.nr_evaluations_
        model_names = [model.name for model in self.models]
        self.history.append_population(
            t, current_eps, population, n_sim, model_names
        )
        logger.debug(
            f"Total samples up to t = {t}: "
            f"{self.history.total_nr_simulations}."
        )

        # acceptance rate and ess
        pop_size = len(population)
        acceptance_rate = pop_size / n_sim
        ess = effective_sample_size(population.get_weighted_distances()['w'])
        logger.info(
            f"Accepted: {pop_size} / {n_sim} = "
            f"{acceptance_rate:.4e}, ESS: {ess:.4e}."
        )

        if self.callback is not None:
            params = self.history_to_params()
            self.callback(params, None, True)

        # prepare next iteration
        self._prepare_next_iteration(
            t=t + 1,
            sample=sample,
            population=population,
            acceptance_rate=acceptance_rate,
        )

        return {
            "successful": True,
            "acceptance_rate": acceptance_rate,
        }

    def history_to_params(self):
        """
        see get_min_dist_parameter in the spike notebook
        """
        history = self.history

        population = history.get_population(history.max_t)
        distance_df = population.get_weighted_distances()
        
        k = np.argmin(distance_df['distance'].values)
        df, w = history.get_distribution(m=0, t=history.max_t)
        return df.iloc[k].values


# coding: utf-8
"""

    NumInverse.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

"""
import numpy as np
from scipy.optimize import minimize

def inverse_at(func, y, x_init):
    def obj_func(p):
        return (y - func(p[0]))**2

    result = minimize(obj_func, np.array([x_init]))
    return result.x



"""
    Selective.AdvancedFrame.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import sys
import numpy as np
try:
    from molass_legacy.KekLib.OurTkinter import Tk, ttk
except:
    import tkinter as Tk
    import tkinter.ttk as ttk
from molass_legacy._MOLASS.Version import is_developing_version

ALLOW_TRADITIONAL = is_developing_version()
USE_COMBOBOX_FOR_PROPS_TEXT = True
MAX_NUM_COMPONENTS = 7
CHANGE_NAMES_DICT = {"CNC": "Components Change",
                     "PSD": "PSD Estimate",
                     "RDR": "RDR Minimization",
                     "KPD": "KPD Minimization",
                     "VPA": "VP Analysis",
                     }

def get_change_name(change_id):
    return CHANGE_NAMES_DICT[change_id]

class AdvancedFrame(Tk.LabelFrame):
    def __init__(self, parent, editor, **kwargs):
        from molass_legacy.KekLib.SaferSpinbox import SaferSpinbox as Spinbox
        from molass_legacy.Selective.Proportions import Proportions_
        global PROPORTIONS_
        PROPORTIONS_ = Proportions_()

        self.parent = parent
        self.editor = editor
        Tk.LabelFrame.__init__(self, parent, text="Advanced Optimization", labelanchor=Tk.N, **kwargs)

        iframe = Tk.Frame(self)
        iframe.pack(padx=0, pady=10)
        irow = 0

        params_array = editor.get_current_params_array()
        num_components = len(params_array)

        nc_frame = Tk.Frame(iframe)
        nc_frame.grid(row=irow, column=0, columnspan=8, padx=10, sticky=Tk.W)
        label = Tk.Label(nc_frame, text="Num Components", state=Tk.DISABLED)
        label.grid(row=0, column=0, sticky=Tk.E)
        self.num_components_label = label
        self.num_components = Tk.IntVar()
        self.num_components.set(num_components)
        sb = Spinbox(nc_frame,
                    textvariable=self.num_components, from_=2, to=MAX_NUM_COMPONENTS,
                    increment=1,
                    justify=Tk.CENTER, width=3, state=Tk.DISABLED )
        sb.grid(row=0, column=1, sticky=Tk.W, padx=5)
        self.num_components_sb = sb
        self.num_components.trace_add('write', self.num_components_tracer)

        self.with_props = Tk.IntVar()
        self.with_props.set(0)
        w = Tk.Checkbutton(nc_frame, text="with", variable=self.with_props)
        w.grid(row=0, column=2, padx=5, pady=5)

        self.with_props_text_var = Tk.StringVar()
        if USE_COMBOBOX_FOR_PROPS_TEXT:
            w = ttk.Combobox(nc_frame, textvariable=self.with_props_text_var, width=18, justify=Tk.CENTER)
            w.grid(row=0, column=3, sticky=Tk.W)
            # w.bind("<<ComboboxSelected>>", self.with_props_tracer)
            self.props_text = w
        else:       
            w = Tk.Text(nc_frame, height=1, width=20)
            w.grid(row=0, column=3, sticky=Tk.W)
            self.props_text = w

        self.with_props_tracer()    # let the states of the widgets be consistent
        self.with_props.trace_add('write', self.with_props_tracer)

        w = Tk.Button(nc_frame, text="Change Num Components", width=20, command=self.change_num_components, state=Tk.DISABLED)
        w.grid(row=0, column=4, columnspan=2, padx=10, pady=5)
        self.cnc_button = w

        irow += 1
        minimizer_frame = Tk.Frame(iframe)
        minimizer_frame.grid(row=irow, column=0, columnspan=8, padx=5, pady=5, sticky=Tk.W+Tk.E)
        for i, weight in enumerate([1, 2, 2]):
            minimizer_frame.columnconfigure(i, weight=weight)

        w = Tk.Button(minimizer_frame, text="Estimate PSD", width=12, command=self.estimate_psd, state=Tk.DISABLED)
        w.grid(row=0, column=0, padx=10, pady=5)
        self.psd_button = w

        w = Tk.Button(minimizer_frame, text="Minimize Rg Diff Ratio", width=20, command=self.minimize_rdr, state=Tk.DISABLED)
        w.grid(row=0, column=1, padx=10, pady=5)
        self.rdr_button = w

        w = Tk.Button(minimizer_frame, text="Minimize Kratky Plot Diff", width=20, command=self.minimize_kpd, state=Tk.DISABLED)
        w.grid(row=0, column=2, padx=10, pady=5)
        self.kpd_button = w

        minv, maxv = self.get_min_max_values()
        self.min_val = Tk.DoubleVar()
        self.min_val.set(minv)
        self.max_val = Tk.DoubleVar()
        self.max_val.set(maxv)
        self.num_divs = Tk.IntVar()
        self.num_divs.set(20)

        irow += 1
        w = Tk.Label(iframe, text="Parameter Limits")
        w.grid(row=irow, column=0, columnspan=5, padx=5, pady=0)
        w = Tk.Label(iframe, text="Number of Divisions")
        w.grid(row=irow, column=5, padx=5)

        irow += 1
        w = Tk.Label(iframe, text="Min")
        w.grid(row=irow, column=0, padx=0, pady=0, sticky=Tk.E)
        sb = Spinbox(iframe,
                    textvariable=self.min_val, from_=0, to=10,
                    increment=0.05,
                    justify=Tk.CENTER, width=6 )
        sb.grid(row=irow, column=1, padx=0)
        w = Tk.Label(iframe, text="")
        w.grid(row=irow, column=2, padx=5)
        w = Tk.Label(iframe, text="Max")
        w.grid(row=irow, column=3, padx=0, sticky=Tk.E)
        sb = Spinbox(iframe,
                    textvariable=self.max_val, from_=0, to=10,
                    increment=0.05,
                    justify=Tk.CENTER, width=6 )
        sb.grid(row=irow, column=4, padx=0)
        sb = Spinbox(iframe,
                    textvariable=self.num_divs, from_=10, to=100,
                    increment=1,
                    justify=Tk.CENTER, width=6, state=Tk.DISABLED )
        sb.grid(row=irow, column=5, padx=5)

        irow += 1
        w = Tk.Label(iframe, text="Proportions\nDefinition")
        w.grid(row=irow, column=0, columnspan=2, padx=5, sticky=Tk.E)

        self.code_text = Tk.Text(iframe, width=36, height=3)
        self.code_text.grid(row=irow, column=2, columnspan=5, padx=5, pady=10, sticky=Tk.W)
        self.set_func_code_text()

        w = Tk.Button(iframe, text="VP Analysis", width=12, command=self.show_vp_analysis, state=Tk.DISABLED)
        w.grid(row=1, rowspan=irow, column=6, padx=10, pady=5)
        self.vpa_button = w

        w = Tk.Button(iframe, text="Reset", width=8, command=self.reset_to_initstatus, state=Tk.DISABLED)
        w.grid(row=irow, rowspan=2, column=6, padx=10, pady=5)

        irow += 1
        progress_frame = Tk.Frame(iframe)
        progress_frame.grid(row=irow, column=0, columnspan=8, padx=10, pady=5, sticky=Tk.W + Tk.E)

        progress_label = Tk.Label(progress_frame, text="Progress")
        progress_label.grid(row=0, column=0, sticky=Tk.E)

        self.progress_bar = ttk.Progressbar(progress_frame, orient ="horizontal", length=430, mode="determinate")
        self.progress_bar.grid(row=0, column=1, padx=10, pady=5)

        if sys.executable.find("pythonw") < 0:
            self.popup_menu = None
            widget = self.parent if self.editor is None else self.editor
            widget.bind("<Button-3>", self.on_right_click)

        self.advanced_widgets = [self.num_components_label, self.num_components_sb, self.cnc_button, self.psd_button]
        self.traditional_widgets = [self.rdr_button, self.kpd_button, self.vpa_button]
        self.after(100, self.enable_buttons)
        self.last_change_id = None  # will be set to one of CHANGE_IDS
        self.change_button_dict = {
            "CNC" : self.cnc_button,
            "PSD" : self.psd_button,
            "RDR" : self.rdr_button,
            "KPD" : self.kpd_button,
            "VPA" : self.vpa_button,
        }

    def get_last_change_id(self):
        return self.last_change_id

    def set_func_code_text(self):
        n = self.num_components.get()
        func_code = PROPORTIONS_.generate_funccode(n)
        self.code_text.delete(1.0, Tk.END)
        self.code_text.insert(1.0, func_code)
        opt_props_text = PROPORTIONS_.get_optimal_props_text(n)
        if USE_COMBOBOX_FOR_PROPS_TEXT:
            PROPORTIONS_.get_proportions_func(func_code)
            values = PROPORTIONS_.get_with_props_values(n)
            self.with_props_text_var.set(values[0])
            self.props_text.config(values=values)
        else:
            self.props_text.config(state=Tk.NORMAL)
            self.props_text.delete(1.0, Tk.END)
            self.props_text.insert(1.0, opt_props_text)
            self.with_props_tracer()

    def with_props_tracer(self, *args):
        if self.with_props.get() == 1:
            self.props_text.config(state=Tk.NORMAL)
        else:
            self.props_text.config(state=Tk.DISABLED)

    def get_min_max_values(self):
        n = self.num_components.get()
        return PROPORTIONS_.get_min_max_values(n)
    
    def num_components_tracer(self, *args):
        minv, maxv = self.get_min_max_values()
        self.min_val.set(minv)
        self.max_val.set(maxv)
        self.num_components_sb.config(state=Tk.NORMAL)
        self.set_func_code_text()

    def get_current_model(self):
        return self.editor.get_current_model()

    def disable_buttons(self):
        for buton in self.advanced_widgets + self.traditional_widgets:
            buton.config(state=Tk.DISABLED)
        self.update()

    def enable_buttons(self):
        widegts = []
        widegts += self.traditional_widgets
        model = self.get_current_model()
        if model.is_traditional():
            allow = ALLOW_TRADITIONAL
        else:
            allow = True
        if allow:
            widegts += self.advanced_widgets
        for w in widegts:
            w.config(state=Tk.NORMAL)
        self.update()

    def update_mode_dependent_state(self):
        self.disable_buttons()
        self.enable_buttons()

    def update_button_status(self, change_id=None):
        if change_id is not None:
            if self.last_change_id is not None:
                self.change_button_dict[self.last_change_id].config(bg="SystemButtonFace", fg="black")
            self.last_change_id = change_id
        if self.last_change_id is not None:
            self.change_button_dict[self.last_change_id].config(bg="green", fg="white")

    def get_prop_vector(self):
        return np.linspace(self.min_val.get(), self.max_val.get(), self.num_divs.get())

    def progress_update(self, progress_value):
        if progress_value >= 0:
            value = progress_value
        else:
            value = self.progress_bar['maximum']
        self.progress_bar['value'] = value
        self.update()

    def show_vp_analysis(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.VariedPropAnalysis
            reload(Selective.VariedPropAnalysis)
        from Selective.VariedPropAnalysis import show_vp_analysis_impl
        prop_func = self.get_proportions_func()
        self.execute_and_monitor(show_vp_analysis_impl, self.num_divs.get(), prop_func=prop_func)

    def execute_and_monitor(self, impl_func, max_progress, **kwargs):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        ret = MessageBox.askokcancel("Confirmation",
                "This may take a few minutes.\n"
                + "Are you sure to proceed?",
                parent=self.parent)
        if not ret:
            return

        prop_func = kwargs.pop('prop_func', None)

        self.disable_buttons()
        self.progress_bar['maximum'] = max_progress
        self.progress_bar['value'] = 0
        self.update()

        if prop_func is not None:
            kwargs['prop_func'] = prop_func

        impl_func(self, **kwargs)
        self.enable_buttons()

    def reset_to_initstatus(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        ret = MessageBox.askokcancel("Confirmation",
                "This will reset the current decomposition to the initial state.\n"
                + "Are you sure to proceed?",
                parent=self.parent)
        if not ret:
            return

        self.editor.reset_current_frame()

    def on_right_click(self, event, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.KekLib.PopupMenuUtils
            reload(KekLib.PopupMenuUtils)
        from molass_legacy.KekLib.PopupMenuUtils import post_popup_menu
        self.create_popup_menu()
        post_popup_menu(self.popup_menu, self.parent, event)

    def create_popup_menu(self):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0)
            self.popup_menu.add_command(label='Estimate Lognormal PSD with PDF', command=self.estimate_lognormal_psd)
            self.popup_menu.add_command(label='Estimate Lognormal PSD with CF', command=lambda: self.estimate_lognormal_psd(with_cf=True))
            self.popup_menu.add_command(label='CF Optimization Study', command=self.cf_optimization_study)
            self.popup_menu.add_command(label='Test Lognormalpore', command=self.test_lognormalpore)
            self.popup_menu.add_command(label='Test Tripore', command=self.test_tripore)

    def get_proportions_func(self, debug=False):
        code_text = self.code_text.get("1.0", Tk.END)
        return PROPORTIONS_.get_proportions_func(code_text, debug=debug)

    def change_num_components(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.NumComponents
            reload(Selective.NumComponents)
        from Selective.NumComponents import change_num_components_impl
        if self.with_props.get() == 1:
            proportions = self.get_proportions_func()
            locals()['proportions'] = proportions      # this will define the proportions function which may be called in the eval below
            if USE_COMBOBOX_FOR_PROPS_TEXT:
                props_text = self.with_props_text_var.get()
            else:
                props_text = self.props_text.get("1.0", Tk.END)
            target_props = eval(props_text)
            if devel:
                print("props_text=", props_text)
                print("target_props=", target_props)
        else:
            target_props = None
        change_num_components_impl(self, target_props=target_props)
        self.update_button_status()

    def minimize_rdr(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.RdrMinimizer
            reload(Selective.RdrMinimizer)
        from Selective.RdrMinimizer import try_rdr_minimization
        self.execute_and_monitor(try_rdr_minimization, 20)
    
    def minimize_kpd(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.KpdMinimizer
            reload(Selective.KpdMinimizer)
        from Selective.KpdMinimizer import try_kpd_minimization
        self.execute_and_monitor(try_kpd_minimization, 20)

    def estimate_psd(self, devel=True):
        if devel:
            from importlib import reload
            import Models.Stochastic.TriporeColumn
            reload(Models.Stochastic.TriporeColumn)
        from molass_legacy.Models.Stochastic.TriporeColumn import estimate_psd_impl
        self.execute_and_monitor(estimate_psd_impl, 10)

    def estimate_lognormal_psd(self, with_cf=False, devel=True):
        if devel:
            from importlib import reload
            import Models.Stochastic.LognormalPoreColumn
            reload(Models.Stochastic.LognormalPoreColumn)
        from molass_legacy.Models.Stochastic.LognormalPoreColumn import estimate_lognormal_psd_impl
        self.execute_and_monitor(estimate_lognormal_psd_impl, 20, with_cf=with_cf)

    def cf_optimization_study(self, devel=True):
        if devel:
            from importlib import reload
            import CharFunc.CfOptimizationStudy
            reload(CharFunc.CfOptimizationStudy)
        from CharFunc.CfOptimizationStudy import cf_optimization_study_impl
        self.execute_and_monitor(cf_optimization_study_impl, 20)

    def test_lognormalpore(self, devel=True):
        if devel:
            from importlib import reload
            import Models.Stochastic.LognormalPore
            reload(Models.Stochastic.LognormalPore)
        from molass_legacy.Models.Stochastic.LognormalPore import lognormalpore_test_impl
        self.execute_and_monitor(lognormalpore_test_impl, 20)

    def test_tripore(self, devel=True):
        if devel:
            from importlib import reload
            import Models.Stochastic.Tripore
            reload(Models.Stochastic.Tripore)
        from molass_legacy.Models.Stochastic.Tripore import tripore_test_impl
        self.execute_and_monitor(tripore_test_impl, 20)

    def test_it(self):
        from importlib import reload
        import Models.Stochastic.Tripore
        reload(Models.Stochastic.Tripore)
        from molass_legacy.Models.Stochastic.Tripore import test_it_from_editor_frame
        print("test it")
        test_it_from_editor_frame(self.editor)

if __name__ == '__main__':
    import sys
    import os
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    context = {}
    func_code = ("def proportions(p):\n"
                 "    return p, 1-p")
    exec(func_code, context)
    proportions = context['proportions']
    print(proportions(0.2))

    root = Tk.Tk()

    class DummyModel:
        def __init__(self):
            pass
        def is_traditional(self):
            return False

        def get_name(self):
            return "EGHA"

    class DummyEditor:
        def __init__(self):
            pass
        def get_current_params_array(self):
            return np.zeros((2,3))
        def bind(self, *args):
            pass
        def get_current_model(self):
            return DummyModel()

    frame = AdvancedFrame(root, DummyEditor(), bd=3, relief=Tk.RIDGE)
    frame.pack()

    root.mainloop()



"""
    AtsasTools.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurSubprocess import exec_subprocess
from molass_legacy.ATSAS.AutoRg import autorg_exe_array
from .IntensityData import IntensityData
from .Guinier import Guinier
from .Result import Result
from .ErrorResult import ErrorResult

def autorg(file, exe_index=0, out_file=None, smaxrg=1.3, verbose=False):
    # print( 'autorg: file=', file )
    assert( os.path.exists( file ) )
    if len(autorg_exe_array) == 0:
        return ErrorResult(), None

    exe_ = autorg_exe_array[ exe_index ]
    opts = ['--smaxrg', str(smaxrg)]

    if out_file is not None:
        opts += ['-o', out_file]

    cmd = [exe_] + opts + [file,  '-f', 'csv']
    out, err = exec_subprocess( cmd, shell=False )
    # print( out, err )
    if err == '':
        lines = out.split( '\n' )
        if len( lines ) >= 3:
            intensity = IntensityData( file )
            x, y, e = intensity.get_guinier_valid_xy()

            File, R, RSD, I, ISD, F, L, Q, A = lines[1].split( ',' )
            if verbose:
                print( R, RSD, I, ISD, F, L, Q, A )
            f, t = int(F)-1, int(L)-1
            f_, t_ = list( map( lambda i: intensity.positive_index[i], ( f, t ) ) )
            Rg = float( R )
            minQ = np.sqrt( x[f_] )
            maxQ = np.sqrt( x[t_] )

            orig_result = Result(
                            type='A',
                            Rg=Rg, Rg_stdev=float( RSD ),
                            I0=float(I), I0_stdev=float(ISD),
                            From=f_, To=t_,
                            Quality=float(Q),
                            Aggregated=float(A),
                            min_qRg=minQ * Rg,
                            max_qRg=maxQ * Rg,
                            min_curvature=None,
                            max_curvature=None,
                            )
            try:
                guinier = Guinier( x, y, e, positive_ratio=intensity.positive_ratio )
                guinier.estimate_rg( [ f_, t_ ], [ 0, 1.3 ] )
                eval_result = guinier.get_result()
            except:
                # as in 20180204/Open01/Open01_00171_sub.dat
                eval_result = None

            return orig_result, eval_result
        else:
            return ErrorResult(), None

    else:
        found = err.find("Data quality")
        if verbose or found < 0:
            print(err)
        return ErrorResult(), None



"""
    Stochastic.ColumnSliceStates.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

def draw_slice_states_impl(fig, ax, grains, pxv, pyv, inmobile_states, debug=False):
    print("draw_slice_states_impl")
 
    dy = 0.01

    maxy = np.max(pyv) + dy
    miny = np.min(pyv) - dy
    print("maxy, miny=",maxy, miny)

    hist, bin_edges = np.histogram(pyv, bins=20)
    print("hist, bin_edges=",hist, bin_edges)
    dy = bin_edges[1] - bin_edges[0]

    def draw_impl(ax):
        ax.set_ylim(0, 1)
        ax.set_xlim(0, np.max(hist)*1.1)
        starts = np.zeros(len(hist))
        height = (bin_edges[1] - bin_edges[0])*0.9
        for label, pyv_ in [('Mobile Phase', pyv[inmobile_states]), ('Stationary Phase', pyv[~inmobile_states])]:
            hist_, _ = np.histogram(pyv_, bins=bin_edges)
            ax.barh(bin_edges[:-1] + dy/2, hist_, left=starts, height=height, label=label)
            starts += hist_
        ax.legend(fontsize=16)

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig_, ax_ = plt.subplots(figsize=(8,20))
            draw_impl(ax_)
            fig_.tight_layout()
            plt.show()
    else:
        ax.cla()
        draw_impl(ax)
        fig.canvas.draw()


# coding: utf-8
"""
    Rgg.RggMixtureModel.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize
from molass_legacy.KekLib.SciPyCookbook import smooth
from Prob.GaussianMixture import gaussian_pdf
from molass_legacy.Peaks.ElutionModels import egh_pdf, compute_egh_params
from pomegranate import *
from .RggEghDistribution import RggEghDistribution
from Rgg.RggUtils import convert_to_probabilitic_data
from .Callbacks import VisualizationtCallback
import molass_legacy.KekLib.DebugPlot as plt

USE_RG = True
if not USE_RG:
    from Prob.GaussianMixture import hist_to_source

USE_GAUSSIAN_MIXTURE_MODEL = False
if USE_GAUSSIAN_MIXTURE_MODEL:
    from sklearn.mixture import GaussianMixture

def sort_dist_params_lists(distributions, weights):
        components = sorted(list(zip(distributions, weights)), key=lambda x: x[0].parameters[0])
        distributions, weights = zip(*components)
        return distributions, weights

class FitResult:
    def __init__(self, model):
        if USE_GAUSSIAN_MIXTURE_MODEL:
            """
            verify the differences of model parameters between 2D (USE_RG) and 1D (not USE_RG)
            """ 
            # print(dir(model))
            # print("model.means_=", model.means_)
            # print("model.covariances_", model.covariances_)
            self.weights = model.weights_
            if USE_RG:
                self.params_list = [(mu, std) for mu, std in model.means_]
            else:
                self.params_list = [(mu[0], np.sqrt(var)) for mu, var in zip(model.means_, model.covariances_)]
        else:
            distributions = model.distributions
            weights = np.exp(model.weights)     # note that weights have been converted to log
            distributions, weights = sort_dist_params_lists(distributions, weights)
            self.weights = weights
            self.model = model = GeneralMixtureModel(distributions, weights)
            self.params_list = [d.parameters for d in distributions]

    def get_params_for_refiner(self):
        ret_params = []
        for w, params in zip(self.weights, self.params_list):
            ret_params.append((w, *params[0:4]))
        return ret_params

class RggMixtureModel:
    def __init__(self, X, num_components=None):
        self.logger = logging.getLogger(__name__)
        self.X = X
        self.num_components = num_components
        if USE_GAUSSIAN_MIXTURE_MODEL:
            cv_type = "spherical"
            self.model = GaussianMixture(n_components=num_components, covariance_type=cv_type)

    def fit(self, init_result=None, peaks=None):
        for k in range(100):
            try:
                return self.fit_impl(init_result=init_result, peaks=peaks)
            except StopIteration as exc:
                # this does not seem to work due to
                # the internal exception handling of GeneralMixtureModel
                print([k], "retrying fit")

    def fit_impl(self, init_result=None, peaks=None):
        if USE_GAUSSIAN_MIXTURE_MODEL:
            self.model.fit(self.X)
            return FitResult(self.model)

        model = None
        if peaks is None:
            if init_result is None:
                model = GeneralMixtureModel.from_samples(RggEghDistribution, self.num_components, self.X)
            else:
                distributions = init_result.model.distributions
                weights = init_result.model.weights
        else:
            distributions, weights = self.get_init_params_from_peaks(peaks)

        if model is None:
            model = GeneralMixtureModel(distributions, weights)

        _ = model.fit(self.X, callbacks=[VisualizationtCallback()])
        return FitResult(model)

    def get_components(self, fit_result, x, y, simple_scale=False):
        ty = np.zeros(len(x))
        cy_list = []
        for w, params in zip(np.exp(fit_result.weights), fit_result.params_list):
            if USE_GAUSSIAN_MIXTURE_MODEL:
                mu, sigma = params
                cy = w * gaussian_pdf(x, mu, sigma)
            else:
                tR, sigma, tau, rg, _ = params
                print(tR, sigma, tau, rg)
                cy = w * egh_pdf(x, tR, sigma, tau)
            cy_list.append(cy)
            ty += cy

        scale = np.max(y)/np.max(ty)
        if simple_scale:
            return [scale*cy for cy in cy_list], scale*ty

        y_positive = y.copy()
        y_positive[y < 0] = 0

        cy_matrix = np.array(cy_list).T

        def obj_func(scales):
            scaled_matrix = cy_matrix * scales
            if False:
                plt.push()
                fig, ax = plt.subplots()
                ax.plot(x, y)
                ty = np.zeros(len(x))
                for cy in scaled_matrix.T:
                    ax.plot(x, cy, ':')
                    ty += cy
                ax.plot(x, ty, ':', color='red')
                fig.tight_layout()
                plt.show()
                plt.pop()
            return np.sum((np.sum(scaled_matrix, axis=1) - y_positive)**2)

        init_scales = np.ones(len(cy_list)) * scale

        result = minimize(obj_func, init_scales)

        ret_ty = np.zeros(len(x))
        ret_cy_list = []
        for scale, cy in zip(result.x, cy_list):
            cy_ = scale*cy
            ret_cy_list.append(cy_)
            ret_ty += cy_

        return ret_cy_list, ret_ty

    def get_init_params_from_peaks(self, rb_peaks):
        distributions = []
        heights = []
        x = rb_peaks.x
        y = rb_peaks.y
        peaks = rb_peaks.get_peaks()
        for peak in peaks:
            ls, pt, rs = peak
            tR = x[pt]
            sigma = (x[pt] - x[ls] + x[rs] - x[pt])/2
            tau = 0
            dist = EghDistribution(tR, sigma, tau)
            distributions.append(dist)
            heights.append(y[pt])

        distributions, weights = sort_dist_params_lists(distributions, heights)
        weights = np.array(weights)
        weights /= np.sum(weights)

        num_peaks = len(peaks)
        num_components = self.num_components
        num_sub_peaks = num_components - num_peaks
        assert num_sub_peaks >= 0
        if num_sub_peaks == 0:
            return distributions, weights

        spline = UnivariateSpline(x, y, s=0, ext=3)
        model = GeneralMixtureModel(distributions, weights)
        model.fit(self.X)

        taus = [d.parameters[2] for d in model.distributions]
        if abs(taus[0]) > abs(taus[-1]):
            sides = [-1, 1]
        else:
            sides = [1, -1]

        distributions = list(distributions)

        for side in sides:
            if side < 0:
                # left side
                i = 0
                dist = model.distributions[i]
                parameters = dist.parameters
                peak_tR,peak_sigma, peak_tau = parameters
                tR = peak_tR - peak_sigma*2.5
                sigma = peak_sigma/4
                tau = peak_tau/4
            else:
                # right side
                i = -1
                dist = model.distributions[i]
                parameters = dist.parameters
                peak_tR,peak_sigma, peak_tau = parameters
                tR = peak_tR + peak_sigma*2.5
                sigma = peak_sigma/4
                tau = peak_tau/4
            h = spline(tR)
            sub_dist = EghDistribution(tR, sigma, tau)
            distributions.insert(i, sub_dist)
            heights.insert(i, h)
            num_peaks += 1
            if num_peaks == num_components:
                break

        distributions, weights = sort_dist_params_lists(distributions, heights)
        weights = np.array(weights)
        weights /= np.sum(weights)

        return distributions, weights



"""
    V2PropOptimizer.PropOptimizerSTC.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from time import sleep



def compute_optimal_proportion(progress_queue, job_args):
    for i in range(10):
        print([i])
        sleep(1)
        progress_queue.put([i])
    progress_queue.put([-1])



# coding: utf-8
"""
    AqBqSimilarity.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.ticker as ticker
from matplotlib.widgets import Button
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

@ticker.FuncFormatter
def major_formatter(x, pos):
    apparant = False if pos is None else (pos - 1)%2 == 0
    return "%.0f" % x if apparant else ""

class AqBqSimilarityDialog(Dialog):
    def __init__( self, parent, dialog):
        self.parent = parent
        self.dialog = dialog
        self.q = dialog.q
        self.A2 = A2 = dialog.A2
        B2 = dialog.B2
        q_slice = slice(dialog.q_index, dialog.q_index2)
        scale = np.average(A2[q_slice])/np.average(B2[q_slice])
        self.B2 = B2*scale

        Dialog.__init__( self, parent, "A(a)-B(q) Similarity", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(14,7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        gs = GridSpec(3,2)
        ax11 = fig.add_subplot(gs[0:2,0])
        ax21 = fig.add_subplot(gs[2,0])
        ax12 = fig.add_subplot(gs[0:2,1])
        ax22 = fig.add_subplot(gs[2,1])

        fig.suptitle("A(q)-B(q) Similarity Inspection", fontsize=20)
        self.draw_curves(ax11, ax21, ax12, ax22)

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        self.mpl_canvas.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=padx, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_curves(self, ax11, ax21, ax12, ax22):
        ax11.set_title("Scaled Log Plot", fontsize=16)
        ax11.set_yscale('log')
        ax11.plot(self.q, self.A2, label='A(q)', color='C1')
        ax11.plot(self.q, self.B2, label='scaled B(q)', color='pink', alpha=0.5)
        ax11.axes.get_xaxis().set_ticks([])
        ax11.legend()

        ax21.plot(self.q, np.log10(self.A2) - np.log10(self.B2), label='log(A(q)) - log(scaled B(q))')
        ax21.legend()

        ax12.set_title("Scaled Linear Plot", fontsize=16)
        ax12.plot(self.q, self.A2, label='A(q)', color='C1')
        ax12.plot(self.q, self.B2, label='scaled B(q)', color='pink', alpha=0.5)
        ax12.axes.get_xaxis().set_ticks([])
        ax12.legend()

        y = self.dialog.compute_min_norm_diff_curve(self.A2, self.B2)
        ax22.plot(self.q, y, label='A(q) - B(q)')
        ax22.legend()

        y_ = y[0:self.dialog.q_limit]
        cd_score = np.sqrt(np.average(y_**2))/self.dialog.scale
        self.dialog.draw_score_text(ax22, y[-1], cd_score*100)



# coding: utf-8
"""
    AffineDemo.py

    Copyright (c) 2018,2024, Masatsuyo Takahashi, KEK-PF
"""

import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.patches     import Polygon
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.Affine                 import Affine


def demo_plot( fig, src_points, tgt_points, with_proof=False ):
    affine  = Affine( src_points, tgt_points )

    src_polygon = Polygon( src_points, alpha=0.2 )
    tgt_polygon = Polygon( tgt_points, alpha=0.2 )

    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )

    ax1.set_title( 'source shapes (on the standard mapping plane)' )
    ax2.set_title( 'affine-transformed shapes (on another Q-plane )' )

    for ax in [ ax1, ax2 ]:
        ax.set_xlim( 0, 1 )
        ax.set_ylim( 0, 1 )

    ax1.add_patch(src_polygon)

    x   = np.arange( 0.1, 0.95, 0.1 )
    src_line    = x * 0.1 + 0.05
    ax1.plot( x, src_line, color='red' )

    ax2.add_patch(tgt_polygon)

    x_, y_ = affine.transform( x, src_line )
    ax2.plot( x_, y_, color='red' )

    for point in src_points:
        x, y = point
        ax1.plot( x, y, 'o', markersize=10 )

    for point in tgt_points:
        x, y = point
        ax2.plot( x, y, 'o', markersize=10 )

    if with_proof:
        if True:
            src_x = [ p[0] for p in src_points ]
            src_y = [ p[1] for p in src_points ]
            tgt_x, tgt_y = affine.transform( src_x, src_y )
            ax2.plot( tgt_x, tgt_y, 'o', color='black', markersize=10 )
        else:
            tgt_p = affine.transform_list(src_points)
            for point in tgt_p:
                x, y = point
                ax2.plot( x, y, 'o', color='black', markersize=10 )

    fig.tight_layout()

class AffineDemoDialog( Dialog ):
    def __init__( self, title, parent=None ):
        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = title
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )

    def show( self ):
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__( self, self.parent, self.title_ )
        # TODO: self.resizable(width=False, height=False)

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        figsize = ( 16, 8 ) if is_low_resolution() else ( 20, 10 )
        fig = plt.figure( figsize=figsize )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        self.draw( fig )
        self.parent.config( cursor='' )

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox( self, frame=None ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)

    def draw( self, fig ):
        src_points  = [ ( 0.1, 0.1 ), ( 0.9, 0.1 ), ( 0.5, 0.9 ) ]
        tgt_points  = [ ( 0.1, 0.1 ), ( 0.9, 0.3 ), ( 0.3, 0.7 ) ]
        demo_plot( fig, src_points, tgt_points )



import numpy                as np
import matplotlib.pyplot    as plt
from Intensity              import Intensity

x = np.arange( 0, 0.8, 1e-3 )

# intensity = Intensity( 0.05, 26 )
intensity = Intensity( 0.2, 26 )

y1 = intensity.term1( x )
y2 = intensity.term2( x )

ii = y1 * y2

Y1 = np.log( y1 )
Y_ = np.log( y1 - ii )

X = x**2

fig = plt.figure( figsize=(16,6) )
ax1 = fig.add_subplot( 121 )
ax2 = fig.add_subplot( 122 )

ax1.set_title( 'Intensity plot of a simple model' )
ax2.set_title( 'Intensity plot of a simple model zoomed in a small-q region' )

for ax in [ax1, ax2]:
    ax.plot( X, Y_, color='red' )
    ax.plot( X, Y1, ':', color='blue' )
    ax.set_xlabel( 'Q' )
    ax.set_ylabel( 'Intensity' )

xmin1, xmax1 = ax1.get_xlim()
ymin1, ymax1 = ax1.get_ylim()

xmin2 = xmin1
xmax2 = xmax1/32

ax2.set_xlim( xmin2, xmax2 )
zoom_size = x.shape[0] // 32

# YP = np.hstack( [ Y1[y1 > 0][0::zoom_size],  Y_[y1 - ii > 0][0:zoom_size] ] )
YP = Y1[y1 > 0][0::zoom_size]

yminp = np.min( YP )
ymaxp = np.max( YP )
ymin2 = (yminp + ymaxp)/2
ymax2 = ymaxp
ax2.set_ylim( ymin2, ymax2 )

ax1.plot(   [ xmin2, xmax2, xmax2, xmin2, xmin2 ],
            [ ymin2, ymin2, ymax2, ymax2, ymin2 ], ':', color='black' )

plt.tight_layout()
plt.show()



# coding: utf-8
"""
    AbsorbanceViewer.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
import os
import numpy                as np
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
from matplotlib.widgets     import Button
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings         import get_setting
from molass_legacy.KekLib.TkUtils                import split_geometry
from molass_legacy.KekLib.TkCustomWidgets        import FileEntry
from .AbsorbanceViewUtils import draw_3d
import matplotlib
USE_CANVAS_3D = True
if USE_CANVAS_3D:
    from Canvas3D import Canvas3D

DEBUG = False

class AbsorbanceViewer( Dialog ):
    def __init__( self, absorbance, helper_info=None ):
        self.grab = 'local'     # used in grab_set
        self.absorbance     = absorbance
        self.helper_info    = helper_info
        self.wl_vector      = absorbance.wl_vector
        self.wvlen_lower    = 245
        self.wvlen_upper    = 450
        if True:
            f = 0
            t = len(absorbance.wl_vector)
        else:
            f = bisect_right( absorbance.wl_vector, self.wvlen_lower )
            t = bisect_right( absorbance.wl_vector, self.wvlen_upper )
        self.i_slice        = slice( f, t )
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

    def show( self, parent, button_labels=[ "OK", "Cancel" ], toolbar=False ):
        assert( len(button_labels) == 2 )
        self.button_labels  = button_labels
        self.toolbar    = toolbar

        Dialog.__init__( self, parent, "AbsorbanceViewer" )
        # TODO: self.resizable(width=False, height=False)

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        figsize = ( 9, 8 ) if is_low_resolution() else ( 11, 10 )
        self.fig = fig = plt.figure( figsize=figsize )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that add_subplot should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        if USE_CANVAS_3D:
            c3d = Canvas3D(fig, 23, 5 )
            ax1 = c3d.get_axis()
            lpm_demo_ax = plt.axes([0.82, 0.26, 0.12, 0.04])
            self.lpm_demo_btn = Button(lpm_demo_ax, 'LPM area', hovercolor='0.975')
            self.lpm_demo_btn.on_clicked(self.toggle_lpm_area)
            self.show_lpm_area = False
            baseplane_ax = plt.axes([0.82, 0.22, 0.12, 0.04])
            self.baseplane_btn = Button(baseplane_ax, 'base plane', hovercolor='0.975')
            self.baseplane_btn.on_clicked(self.toggle_baseplane)
            self.show_baseplane = False
        else:
            ax1 = fig.add_subplot( 111, projection='3d' )
            # ax2 = fig.add_subplot( 122 )
        self.axes   = [ax1, None]

        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()

        self.draw()
        self.parent.config( cursor='' )

        if self.toolbar:
            self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
            self.toolbar.update()

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)
        self.button_frame = box

        w = Tk.Button(box, text=self.button_labels[0], width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text="Save", width=10, command=self.save_dialog)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.save_button = w
        w = Tk.Button(box, text=self.button_labels[1], width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def ok( self, *args ):
        self.applied  = True
        self.destroy()

    def cancel( self, *args ):
        self.applied  = False
        self.destroy()

    def draw( self ):
        ax1 = self.axes[0]

        draw_3d( ax1, self.absorbance, self.wvlen_lower, self.wvlen_upper, self.i_slice, low_percentile=self.show_lpm_area )

        if USE_CANVAS_3D:
            if self.show_baseplane:
                self.draw_baseplane(ax1)
        else:
            self.fig.tight_layout()

        self.mpl_canvas.draw()

    def toggle_lpm_area( self, event ):
        # print('toggle_lpm_area')
        self.show_lpm_area ^= True
        ax1 = self.axes[0]
        ax1.cla()
        self.draw()

    def toggle_baseplane( self, event ):
        # print('toggle_baseplane')
        self.show_baseplane ^= True
        ax1 = self.axes[0]
        ax1.cla()
        self.draw()

    def draw_baseplane( self, ax ):
        A, B, C = self.absorbance.get_baseplane_params()
        # print( 'draw_baseplane(%d): A=%.4g, B=%.4g, C=%.4g' % ( id(self.absorbance), A, B, C ) )
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        xx, yy = np.meshgrid( np.linspace(xmin, xmax, 10), np.linspace(ymin, ymax, 10) )
        zz = xx * A + yy * B + C
        ax.plot_surface(xx, yy, zz, color='red', alpha=0.1 )

    def save_dialog( self ):
        from .AbsorbanceSaveDialog import AbsorbanceSaveDialog
        dialog = AbsorbanceSaveDialog( self )
        dialog.show()



# coding: utf-8
"""
    WlsDecomposer.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import statsmodels.api      as sm
import matplotlib.pyplot    as plt

DEBUG   = False

class WlsDecomposer:
    def __init__( self, ecurve, data, drift=False, in_folder=None ):
        self.data   = data

        peaks_ = [ int(info[1]) for info in ecurve.peak_info ]
        if drift:
            peaks_ += [-1]

        print( peaks_ )
        A = data[:,peaks_]
        print( A.shape )

        params_list = []
        for j in range( self.data.shape[1] ):
            params = self.solve_an_elution( A, j )
            params_list.append( params )

        solution = np.array( params_list ).T
        for k in range( solution.shape[0] ):
            plt.plot( solution[k,:], label='component' + str(k) )

        plt.legend()
        plt.tight_layout()
        plt.show()

        if drift:
            A[:, -1] *= 10

        Apinv = np.linalg.pinv( A )
        print( Apinv.shape )
        C   = np.dot( Apinv, data )

        if in_folder is not None:
            plt.title( in_folder )

        num_components   = len( peaks_ )
        for k in range( num_components ):
            plt.plot( C[k,:], label='component' + str(k) )

        plt.legend()
        plt.tight_layout()
        plt.show()

    def solve_an_elution( self, A, j ):
        y   = self.data[:,j]
        x   = A
        X   = sm.add_constant(x)
        mod = sm.OLS( y, X )
        res = mod.fit()
        # print( 'res.params=', res.params )
        return res.params



# coding: utf-8
"""
    CdNoiseMix.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Button
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.Models.ElutionCurveModels import emg_x_from_height_ratio
from SvdDenoise import get_denoised_data
from .SimData import SimData

class CdNoiseMixDialog(Dialog):
    def __init__( self, parent, seed=None, show_rank=False, for_paper=False):
        in_folder = 'D:/PyTools/reports/analysis-erh/preview_results'
        a_file = in_folder + '/A1.dat'
        b_file = in_folder + '/B1.dat'
        self.simd = simd = SimData(a_file, b_file)
        self.q = simd.q
        h, mu, sigma, tau = simd.e_params[0]
        f, t = [int(p+0.5) for p in emg_x_from_height_ratio(0.5, mu, sigma, tau)]
        print("(f, t)=", (f, t))
        self.eslice = slice(f,t+1)
        if seed is None:
            seed = np.random.randint(1000, 9999)
        self.seed = seed
        self.show_rank = show_rank
        self.for_paper = for_paper
        np.random.seed(seed)
        Dialog.__init__( self, parent, "CdNoiseMixDialog", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(23, 12))
        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        fig.suptitle("Apparent B(q)'s at various levels of Noises and SCD's; random seed=%d" % self.seed, fontsize=30)

        noise_levels = [0.01, 0.02, 0.05]
        cd_levels = [0.05, 0.1, 1, 2, 5]

        nrows = len(noise_levels)
        ncols = len(cd_levels)
        gs = GridSpec(nrows, ncols)
        axes = []
        for i, noise in enumerate(noise_levels):
            axes_row = []
            for j, cd in enumerate(cd_levels):
                ax = fig.add_subplot(gs[i,j])
                ax.set_title("Noise=%.g, SCD=%.g" % (noise, cd), fontsize=16)
                self.draw_aqbq(ax, noise, cd)
                axes_row.append(ax)
            axes.append(axes_row)
        axes = np.array(axes)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9, bottom=0.05)
        self.mpl_canvas.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(bottom_frame, width=width*0.25)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_aqbq(self, ax, noise, cd):
        sd = self.simd
        i = sd.i
        M, _ = sd.get_data(noise, cd)
        eslice = self.eslice
        c = M[i,eslice]
        c = c/np.max(c)
        C = np.array([c, c**2])
        M_ = get_denoised_data(M[:,eslice], rank=2)
        P_ = M_ @ np.linalg.pinv(C)
        a = P_[:,0]
        b = P_[:,1]

        if self.for_paper:
            aq_color = 'red'
            bq_color = 'blue'
        else:
            aq_color = 'C1'
            bq_color = 'cyan'

        ax.plot(self.q, a, color=aq_color)
        ax.plot(self.q, b, color=bq_color, alpha=0.5)

        if not self.for_paper:
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(self.q, b, color='pink', alpha=0.5)

        if self.show_rank:
            self.draw_exact_rank(ax, self.simd.a, a, M, c)

    def draw_exact_rank(self, ax, truth, a, M, c):
        solution = a
        def obj_func(pv):
            return np.sum((solution*pv[0] - truth)**2)

        res = minimize(obj_func, (1,))
        rank2_score = res.fun

        eslice = self.eslice
        M_ = get_denoised_data(M[:,eslice], rank=1)
        C = np.array([c])
        P_ = M_ @ np.linalg.pinv(C)
        solution = P_[:,0]
        res = minimize(obj_func, (1,))
        rank1_score = res.fun

        print('scores=(%.3g, %.3g)' % (rank1_score, rank2_score))
        rank = 1 if rank1_score < rank2_score else 2

        tx = np.average(ax.get_xlim())
        ty = np.average(ax.get_ylim())

        ax.text(tx, ty, str(rank), fontsize=200, alpha=0.1, color='blue', ha='center', va='center')



# coding: utf-8
"""
    SharedArray.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np

from multiprocessing.shared_memory import SharedMemory

class SharedArray:
    def __init__(self, array=None, tuple_=None, name=None):
        if array is None:
            assert tuple_ is not None
            name, shape, dtype = tuple_
            existing_shm = SharedMemory(name=name)
            sh_array = np.ndarray(shape, dtype=dtype, buffer=existing_shm.buf)
            self.tuple_ = tuple_
            self.shm = existing_shm
            self.sh_array = sh_array
        else:
            shm = SharedMemory(create=True, size=array.nbytes)
            name = shm.name
            sh_array = np.ndarray(array.shape, dtype=array.dtype, buffer=shm.buf)
            sh_array[:] = array[:]  # Copy the original data into shared memory
            self.tuple_ = (name, array.shape, array.dtype)
            self.shm = shm
            self.sh_array = sh_array

    def get_array(self):
        return self.sh_array

    def get_tuple(self):
        return self.tuple_



"""
    Batch.DataBridge.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.Batch.FullBatch import FullBatch

class DataBridge(FullBatch):
    def __init__(self, sd, pre_recog, treat=None):
        self.logger = logging.getLogger(__name__)
        self.pre_recog = pre_recog
        FullBatch.__init__(self)
        if treat is None:
            from SecSaxs.DataTreatment import DataTreatment
            # task: inspect to make these consistent with OptStategyDialog
            correction = 1
            trimming = 2
            treat = DataTreatment(route="v2", trimming=trimming, correction=correction, unified_baseline_type=self.unified_baseline_type)

        # these duplicated procesures are currently required to get base_curve_info correctly
        # i.e., current benefits from using this class are to avoid duplication of loading and pre-recognition
        self.sd = trimmed_sd = treat.get_trimmed_sd(sd, pre_recog)
        self.corrected_sd = treat.get_corrected_sd(sd, pre_recog, trimmed_sd)

        self.base_curve_info = treat.get_base_curve_info()

    def prepare_for_lrf_source(self, num_components=None):
        self.exact_num_peaks = num_components
        uv_x, uv_y, xr_x, xr_y, baselines = self.get_curve_xy(return_baselines=True)
        uv_y_ = uv_y - baselines[0]
        xr_y_ = xr_y - baselines[1]
        self.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_, num_peaks=num_components)
        self.set_lrf_src_args1(uv_x, uv_y, xr_x, xr_y, baselines)



"""
    V2PropOptimizer.PropOptimizerDialogs.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import queue
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText

class PropOptimizerProgress(Dialog):
    def __init__(self, parent, job_args):
        self.done = False
        self.job_args = job_args

        Dialog.__init__(self, parent, "Proportion Optimizer Progress", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        self.progress_bar = ttk.Progressbar(body_frame, orient ="horizontal", length=500, mode="determinate")
        self.progress_bar.pack(padx=20, pady=10)

        self.progress_text = ReadOnlyText(body_frame, width=80, height=5)
        self.progress_text.pack(padx=20, pady=10)
        self.progress_text.update()

        self.after(200, self.start_optimizer)

    def buttonbox(self):
        # task: pack the close button so that it won't hide
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, pady=10)
        self.okcancel_btn = w

    def cancel(self):
        Dialog.cancel(self)

    def start_optimizer(self):
        from molass_legacy.KekLib.KillableThread import Thread
        self.progress_queue = queue.Queue()
        self.thread = Thread(
                        target=self.optimizer_thread,
                        name='V2PropOptimizer',
                        args=[]
                        )
        self.thread.start()
        self.counter = 0
        self.progress_bar['maximum'] = 10
        self.last_index = self.progress_text.index(Tk.INSERT)
        self.after(200, self.update_progress_text)

    def update_progress_text(self):
        progress = self.progress_queue.get()
        print("progress=", progress)
        if progress is not None:
            i = progress[0]
            if i < 0:
                self.progress_bar['value'] = self.progress_bar['maximum']
                self.done = True
                self.thread.join()
                self.after(200, self.cancel)
                return
            else:
                self.progress_bar['value'] = i
                self.update()
        self.after(200, self.update_progress_text)

    def optimizer_thread(self, devel=True):
        if devel:
            from importlib import reload
            import V2PropOptimizer.PropOptimizer
            reload(V2PropOptimizer.PropOptimizer)
        from V2PropOptimizer.PropOptimizerUtils import compute_optimal_proportion
        compute_optimal_proportion(self.progress_queue, self.job_args)

class PropOptimizerResult(Dialog):
    def __init__(self, parent, result):
         Dialog.__init__(self, parent, "Proportion Optimizer Progress", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        text = Tk.Label(body_frame, text="In Preparation")
        text.pack()

class JobArgs:
    def __init__(self, **entries): 
        self.__dict__.update(entries)
    def __str__(self):
        return str(self.__dict__)

def show_optimizer_dialog(parent, modelname, init_prop, v2_optimizer):
    job_args = JobArgs(modelname=modelname, init_prop=init_prop, v2_optimizer=v2_optimizer)
    dialog = PropOptimizerProgress(parent, job_args)
    dialog.show()

    if not dialog.done:
        return

    result = None
    dialog = PropOptimizerResult(parent, result)
    dialog.show()



# coding: utf-8
"""

    ファイル名：   Settings.py

    Copyright (c) 2016-2021, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.PersistentInfo import PersistentInfo
try:
    from molass_legacy._MOLASS.Version        import get_version_string
except:
    from molass_legacy.AutorgKek.AppVersion     import get_com_version_string as get_version_string

# from MachineTypes       import get_display_resolution, get_chassistype_name, get_monitors

# TODO: import wmi into exe
# width, height = get_display_resolution()
# print( 'display width, height=', width, height )

# chassistype_name = get_chassistype_name()
# print( chassistype_name )

# TODO: threading because it takes a significant amountt of time
# for monitor in get_monitors():
#    print( 'monitor=', monitor )
"""
if chassistype_name == 'Desktop':
    default_main_geometry = '1080x750'
else:
    default_main_geometry = '1080x700'
"""
default_main_geometry = '1080x500'
mask_           = None

stamp_name = 'version_stamp'

settings_ = None

ITEM_DEFAULTS = {
    stamp_name              : None,
    'recent_folders'        : {},
    'atsas_exe_paths'       : [],
    'result_book'           : 'analysis_report.xlsx',
    'quality_weighting'     : [
                                0.2,    # basic_quality
                                0.2,    # positive_score
                                0.0,    # fit_consistency
                                0.3,    # rg_stdev_score
                                0.3,    # q_rg_score
                              ],
    }

ALTERNATIVE_WEIGHTS = [ 0.2, 0.2, 0.3, 0.2, 0.1 ]

def reload_settings():
    global setting_info
    global settings_

    setting_info    = PersistentInfo( setting_file )
    settings_       = setting_info.get_dictionary()

def clear_settings():
    global settings_
    settings_ = {}
    setting_info.set_dictionary( settings_ )

def get_setting( item ):
    assert( item in ITEM_DEFAULTS )
    if settings_ is None:
        initialize()
    value = settings_.get( item )
    if value is None:
        value = ITEM_DEFAULTS.get( item )
        set_setting( item, value )
    return value

def set_setting( item, value ):
    assert( item in ITEM_DEFAULTS )
    settings_[ item ] = value

def save_settings():
    setting_info.save()

def check_version_stamp():
    version_stamp = get_setting( stamp_name )
    if version_stamp is None or version_stamp != get_version_string():
        # TODO: MessageBox
        clear_settings()
    set_setting( stamp_name, get_version_string() )

def initialize(alt_folder=None):
    # 永続的な記憶をオブジェクト化する（メモリに取り込む）。
    global setting_info, settings_
    setting_file    = 'settings.dump'
    setting_info    = PersistentInfo( setting_file, alt_folder=alt_folder )
    settings_       = setting_info.get_dictionary()
    check_version_stamp()
    # print("initialize", settings_)

def overriding_init(settings):
    global settings_
    settings_ = settings
